'''
.. autoclass:: TestLoader
   :members:
   :member-order: bysource

'''
import os
import re
import sys
import unittest

import pulsar
from pulsar.utils.importer import import_system_file

from .result import TestRunner


no_tags = set(('tests', 'test'))
build_directories = set(('build', 'dist', 'node_modules'))
test_patterns = [re.compile(r'''test_(?P<name>.*).py'''),
                 re.compile(r'''(?P<name>.*)_test.py'''),
                 re.compile(r'''tests.py''')]


def issubclass_safe(cls, base_cls):
    try:
        return issubclass(cls, base_cls)
    except TypeError:
        return False


def skip_module(mod_name):
    return (mod_name.startswith('_') or
            mod_name.startswith('.') or
            mod_name in build_directories)


class TestLoader:
    '''Classes used by the :class:`.TestSuite` to aggregate tests
    from a list of paths.

    The way it works is simple, you give a *root* directory and a list
    of submodules where to look for tests.

    :parameter root: root path passed by the :class:`.TestSuite`.
    :parameter modules: list (or tuple) of entries where to look for tests.
        Check :ref:`loading test documentation <apps-test-loading>` for
        more information.
    :parameter runner: The :class:`.TestRunner` passed by the test suite.
    '''
    def __init__(self, suite):
        self.stream = pulsar.get_stream(suite.cfg)
        self.runner = TestRunner(suite.cfg.test_plugins, self.stream)
        self.abort_message = self.runner.configure(suite.cfg)
        self.root = suite.root_dir
        self.modules = suite.cfg.test_modules
        self.logger = suite.logger

    def tags(self, include=None, exclude=None):
        """Return a generator of tag information
        """
        for tag, file_path_function in self.test_files(include, exclude):
            module = self.import_module(*file_path_function).module()
            if module:
                doc = module.__doc__
                if doc:
                    tag = '{0} - {1}'.format(tag, doc)
                yield tag

    def import_module(self, file_name, test_function=None):
        return ModuleImporter(file_name, test_function, self.logger)

    def test_files(self, include=None, exclude=None):
        """List of ``tag``, ``modules`` pairs.

        :parameter tags: optional list of tags to include, if not available
            all tags will be included.
        :parameter exclude_tags: optional list of tags to exclude.
            If not provided no tags will be excluded.
        """
        d = dict(self._test_files(include, exclude))
        return [(k, d[k]) for k in sorted(d)]

    # INTERNALS
    def _all_tags(self, tag):
        bits = tag.split('.')
        tag, rest = bits[0], bits[1:]
        yield tag
        for b in rest:
            tag += '.' + b
            yield tag

    def _check_tag(self, tag, import_tags, exclude_tags):
        '''Return ``True`` if ``tag`` is in ``import_tags``.'''
        if exclude_tags:
            alltags = list(self._all_tags(tag))
            for exclude_tag in exclude_tags:
                for bit in alltags:
                    if bit == exclude_tag:
                        return
        if import_tags:
            found = None
            alltags = list(self._all_tags(tag))
            for import_tag in import_tags:
                allitags = list(self._all_tags(import_tag))
                for bit in alltags:
                    if bit == import_tag:
                        return True
                    elif bit in allitags:
                        found = allitags[allitags.index(bit)+1:]
            return found
        else:
            return True

    def _test_files(self, include, exclude):
        if not self.modules:
            yield from self._get_tests(self.root, include, exclude)
        else:
            for name in self.modules:
                absolute_path = os.path.join(self.root, name)
                if os.path.isdir(absolute_path):
                    self.logger.debug('Loading from "%s"', name)
                    yield from self._get_tests(absolute_path, include, exclude)
                else:
                    raise ValueError('%s cannot be found in %s directory.'
                                     % (name, self.root))

    def _get_tests(self, path, include_tags, exclude_tags, tags=None):
        """Collect python modules for testing.

            :parameter path: directory path where to search. Files starting
                with ``_`` or ``.`` are excluded from the search,
                as well as non-python files.
            :parameter dotted_path: the dotted python path equivalent
                of ``path``.
            :parameter parent: the parent module for the current one.
                This parameter is passed by this function recursively
            :return: a generator of tag, module path pairs.
        """
        if tags is None:
            tags = []

        for mod_name in os.listdir(path):
            if skip_module(mod_name):
                continue
            mod_path = os.path.join(path, mod_name)
            is_file = os.path.isfile(mod_path)
            if is_file:
                tag = self._match(mod_name)
                if tag is None:  # does not match and is a file, skip.
                    continue
                m_tags = tags + list(tag)
            else:
                m_tags = tags if mod_name in no_tags else tags + [mod_name]

            tag = '.'.join(m_tags)
            test_function = None

            if tag or is_file:
                found = self._check_tag(tag, include_tags, exclude_tags)
                if not found:
                    continue
                if isinstance(found, list) and is_file:
                    if len(found) > 1:
                        continue
                    test_function = found[0][len(tag)+1:]
                    if not test_function.startswith('test'):
                        continue

            if is_file:
                yield tag, (mod_path, test_function)
            else:
                yield from self._get_tests(mod_path, include_tags,
                                           exclude_tags, m_tags)

    def _match(self, name):
        for pattern in test_patterns:
            p = pattern.search(name)
            if p:
                return p.groups(0)


class ModuleImporter:

    def __init__(self, file_name, test_function, logger):
        self.file_name = file_name
        self.test_function = test_function
        self.logger = logger

    def __iter__(self):
        module = self.module()
        if module:
            test_function = self.test_function
            for name in dir(module):
                obj = getattr(module, name)
                if issubclass_safe(obj, unittest.TestCase):
                    if self.test_function and not hasattr(obj, test_function):
                        continue
                    yield obj, test_function

    def module(self):
        if not hasattr(self, '_module'):
            module = None
            try:
                module = import_system_file(self.file_name, False)
            except ImportError:
                self.logger.error('Failed to import module %s. Skipping.',
                                  self.file_name, exc_info=True)
                self.logger.debug('Full python path:\n%s', '\n'.join(sys.path))
            except Exception:
                self.logger.critical('Failed to import module %s. Skipping.',
                                     self.file_name, exc_info=True)
            self._module = module
        return self._module



import os

from multiprocessing import current_process

try:
    import coverage
except ImportError:
    coverage = None


class Coverage:
    '''Coverage mixin for actors.
    '''
    @property
    def coverage(self):
        return getattr(current_process(), '_coverage', None)

    def start_coverage(self):
        if self.cfg.coverage:
            if not coverage:
                self.logger.error('Coverage module not installed. '
                                  'Cannot start coverage.')
                return
            if self.is_arbiter():
                if not self.coverage:
                    self.logger.warning('Start coverage')
                    p = current_process()
                    p._coverage = coverage.Coverage(data_suffix=True)
                    coverage.process_startup()
                    p._coverage.start()
                config_file = self.coverage.config_file
                os.environ['COVERAGE_PROCESS_START'] = config_file
            elif self.cfg.concurrency == 'process':
                coverage.process_startup()

    def stop_coverage(self):
        cov = self.coverage
        if cov and self.is_arbiter():
            self.logger.warning('Saving coverage file')
            cov.stop()
            cov.save()
            c = coverage.Coverage()
            c.combine()
            c.save()
            self.stream.write(
                'Coverage file available. Type "coverage html" '
                'for a report\n')




cdef extern from "websocket.h":
    object websocket_mask(const char* chunk, const char* key,
                          size_t chunk_length, size_t mask_length)


cdef inline bytes to_bytes(object value, str encoding):
    if isinstance(value, bytes):
        return value
    elif isinstance(value, str):
        return value.encode(encoding)
    else:
        raise TypeError('Requires bytes or string')



import unittest

from pulsar import HttpRedirect


class WsgiRequestTests(unittest.TestCase):

    def test_httpredirect(self):
        location = 'https://pythonhosted.org/pulsar'
        r = HttpRedirect(location)
        self.assertEqual(r.location, location)
        r = HttpRedirect(location, headers=[('Content-Type', 'text/html')])
        self.assertEqual(len(r.headers), 2)
        self.assertEqual(r.location, location)



import re
import unittest

from pulsar.apps.ds import redis_to_py_pattern


class TestUtils(unittest.TestCase):

    def match(self, c, text):
        self.assertEqual(c.match(text).group(), text)

    def not_match(self, c, text):
        self.assertEqual(c.match(text), None)

    def test_redis_to_py_pattern(self):
        p = redis_to_py_pattern('h?llo')
        c = re.compile(p)
        self.match(c, 'hello')
        self.match(c, 'hallo')
        self.not_match(c, 'haallo')
        self.not_match(c, 'hallox')
        #
        p = redis_to_py_pattern('h*llo')
        c = re.compile(p)
        self.match(c, 'hello')
        self.match(c, 'hallo')
        self.match(c, 'hasjdbvhckjcvkfcdfllo')
        self.not_match(c, 'haallox')
        self.not_match(c, 'halloouih')
        #
        p = redis_to_py_pattern('h[ae]llo')
        c = re.compile(p)
        self.match(c, 'hello')
        self.match(c, 'hallo')
        self.not_match(c, 'hollo')



#
# Required by Cython to build Hiredis extensions
#
import os
import sys
from distutils.extension import Extension
from distutils.command.build_ext import build_ext
from distutils.errors import (CCompilerError, DistutilsExecError,
                              DistutilsPlatformError)

path = os.path.join('extensions', 'lib')
ext_file = os.path.join(path, 'lib.c')
ext_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError)

if sys.platform == 'win32':
    # 2.6's distutils.msvc9compiler can raise an IOError when failing to
    # find the compiler
    ext_errors += (IOError,)


class BuildFailed(Exception):

    def __init__(self, msg=None):
        if not msg:
            msg = str(sys.exc_info()[1])
        self.msg = msg


class tolerant_build_ext(build_ext):
    # This class allows C extension building to fail. From SQLAlchemy

    def run(self):
        try:
            if not os.path.isfile(ext_file):
                try:
                    from Cython.Build import cythonize
                except ImportError:
                    raise BuildFailed('Cython not installed') from None
                self.extensions = cythonize(self.extensions,
                                            include_path=[path])
            super().run()
        except DistutilsPlatformError:
            raise BuildFailed from None

    def build_extension(self, ext):
        try:
            super().build_extension(ext)
        except ext_errors:
            raise BuildFailed from None
        except ValueError:
            # this can happen on Windows 64 bit, see Python issue 7511
            if "'path'" in str(sys.exc_info()[1]):  # works with both py 2/3
                raise BuildFailed from None
            raise


def params(cython=False):
    if not cython:
        cython = not os.path.isfile(ext_file)

    if cython and os.path.isfile(ext_file):
        os.remove(ext_file)

    file_name = 'lib.pyx' if cython else 'lib.c'

    extension = Extension('pulsar.utils.lib',
                          [os.path.join(path, file_name)],
                          include_dirs=[path])

    extensions = [extension]

    return {'ext_modules': extensions,
            'cmdclass': {'build_ext': tolerant_build_ext},
            'include_dirs': [path]}



from pulsar.apps.http import HttpClient

from .utils import wait


class GreenHttp:
    """Wraps an http client so we can use it with greenlets
    """
    def __init__(self, http=None):
        self._http = http or HttpClient()

    def __getattr__(self, name):
        return getattr(self._http, name)

    def get(self, url, **kwargs):
        kwargs.setdefault('allow_redirects', True)
        return self.request('GET', url, **kwargs)

    def options(self, url, **kwargs):
        kwargs.setdefault('allow_redirects', True)
        return self.request('OPTIONS', url, **kwargs)

    def head(self, url, **kwargs):
        return self.request('HEAD', url, **kwargs)

    def post(self, url, **kwargs):
        return self.request('POST', url, **kwargs)

    def put(self, url, **kwargs):
        return self.request('PUT', url, **kwargs)

    def patch(self, url, **kwargs):
        return self.request('PATCH', url, **kwargs)

    def delete(self, url, **kwargs):
        return self.request('DELETE', url, **kwargs)

    def request(self, method, url, **kw):
        return wait(self._http.request(method, url, **kw), True)



import unittest

from pulsar.apps.test import dont_run_with_thread

from tests.async.actor import ActorTest


@dont_run_with_thread
class TestActorMultiProcess(ActorTest, unittest.TestCase):
    concurrency = 'process'



import unittest

from pulsar.utils.httpurl import hasextensions, HttpParser

from tests.http import base


@unittest.skipUnless(hasextensions, 'Requires C extensions')
class TestHttpClientWithPythonParser(base.TestHttpClient):

    @classmethod
    def parser(cls):
        return HttpParser



import unittest

from pulsar import send

try:
    from pulsar.apps import greenio
    run_in_greenlet = greenio.run_in_greenlet
except ImportError:
    greenio = None

    def run_in_greenlet(f):
        return f


from examples.echo.manage import server, Echo


class EchoGreen(Echo):
    '''An echo client which uses greenlets to provide implicit
    asynchronous code'''

    def __call__(self, message):
        connection = greenio.wait(self.pool.connect())
        with connection:
            consumer = connection.current_consumer()
            consumer.start(message)
            greenio.wait(consumer.on_finished)
            return consumer if self.full_response else consumer.buffer


@unittest.skipUnless(greenio, "Requires the greenlet module")
class TestGreenIo(unittest.TestCase):
    __benchmark__ = True
    __number__ = 1000

    @classmethod
    async def setUpClass(cls):
        s = server(name=cls.__name__.lower(), bind='127.0.0.1:0')
        cls.server_cfg = await send('arbiter', 'run', s)
        cls.client = Echo(cls.server_cfg.addresses[0])
        cls.green = EchoGreen(cls.server_cfg.addresses[0])
        cls.msg = b'a'*2**13
        cls.pool = greenio.GreenPool()

    async def test_yield_io(self):
        result = await self.client(self.msg)
        self.assertEqual(result, self.msg)

    @run_in_greenlet
    def test_green_io(self):
        result = self.green(self.msg)
        self.assertEqual(result, self.msg)

    async def test_green_pool(self):
        result = await self.pool.submit(self.green, self.msg)
        self.assertEqual(result, self.msg)



from random import randint
import struct
import unittest

from pulsar import ProtocolError, HAS_C_EXTENSIONS
from pulsar.utils.websocket import frame_parser, parse_close


def i2b(args):
    return bytes(bytearray(args))


class FrameTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.bdata = i2b((randint(0, 255) for v in range(256)))
        cls.large_bdata = i2b((randint(0, 255) for v in range(64*1024)))

    def parser(self, pyparser=False, **kw):
        return frame_parser(**kw)

    def test_version(self):
        self.assertRaises(ProtocolError, self.parser, version='bla')

    def test_server(self):
        server = self.parser()
        self.assertEqual(server.decode_mask_length, 4)
        self.assertEqual(server.encode_mask_length, 0)
        self.assertEqual(server.max_payload, 1 << 63)

    def test_both_masked(self):
        server = self.parser(kind=2)
        self.assertEqual(server.decode_mask_length, 0)
        self.assertEqual(server.encode_mask_length, 0)
        server = self.parser(kind=3)
        self.assertEqual(server.decode_mask_length, 4)
        self.assertEqual(server.encode_mask_length, 4)

    def testCloseFrame(self):
        parser = self.parser(kind=2)
        close_message = struct.pack('!H', 1000) + b'OK'
        f = parser.encode(close_message, opcode=0x8)
        self.assertEqual(close_message, f[2:])

    def testControlFrames(self):
        s = self.parser()
        c = self.parser(kind=1)
        #
        chunk = s.close(1001)
        frame = c.decode(chunk)
        self.assertTrue(frame.final)
        self.assertEqual(frame.opcode, 8)
        code, reason = parse_close(frame.body)
        self.assertEqual(code, 1001)
        #
        chunk = s.ping('Hello')
        frame = c.decode(chunk)
        self.assertTrue(frame.final)
        self.assertEqual(frame.opcode, 9)
        self.assertEqual(i2b((0x89, 0x05, 0x48, 0x65, 0x6c, 0x6c, 0x6f)),
                         chunk)
        self.assertEqual(frame.body, b'Hello')
        self.assertRaises(ProtocolError, s.ping, self.bdata)
        #
        chunk = s.pong('Hello')
        frame = c.decode(chunk)
        self.assertTrue(frame.final)
        self.assertEqual(frame.opcode, 10)
        self.assertEqual(frame.body, b'Hello')
        self.assertRaises(ProtocolError, s.pong, self.bdata)

    def test_conntrol_frames_fragmented(self):
        c = self.parser(kind=1)
        for opcode in (8, 9, 10):
            chunk = c.encode('test', opcode=opcode, final=False)
            s = self.parser()
            try:
                s.decode(chunk)
            except ProtocolError:
                pass
            else:
                raise Exception('Protocol error not raised')

    def testUnmaskedDataFrame(self):
        parser = self.parser(kind=2)
        data = parser.encode('Hello')
        f = parser.decode(data)
        self.assertEqual(f.opcode, 1)
        self.assertEqual(len(f.body), 5)
        self.assertFalse(f.masking_key)
        #
        self.assertEqual(i2b((0x81, 0x05, 0x48, 0x65, 0x6c, 0x6c, 0x6f)), data)
        f1 = parser.encode('Hel', final=False)
        f2 = parser.continuation('lo', final=True)
        self.assertEqual(i2b((0x01, 0x03, 0x48, 0x65, 0x6c)), f1)
        self.assertEqual(i2b((0x80, 0x02, 0x6c, 0x6f)), f2)

    def testBinaryDataFrame(self):
        s = self.parser()
        c = self.parser(kind=1)
        #
        chunk = s.encode(self.bdata, opcode=2)
        frame = c.decode(chunk)
        self.assertEqual(frame.opcode, 2)
        self.assertFalse(frame.masking_key)
        self.assertEqual(frame.body, self.bdata)
        self.assertEqual(struct.pack('!BBH', 0x82, 0x7E, 0x0100),
                         chunk[:4])
        #
        chunk = s.encode(self.large_bdata, opcode=2)
        frame = c.decode(chunk)
        self.assertEqual(frame.opcode, 2)
        self.assertFalse(frame.masking_key)
        self.assertEqual(frame.body, self.large_bdata)
        self.assertEqual(struct.pack('!BBQ', 0x82, 0x7F, 0x0000000000010000),
                         chunk[:10])

    def testMaskData(self):
        client = self.parser(kind=1)
        masking_key = i2b((0x37, 0xfa, 0x21, 0x3d))
        chunk = client.encode('Hello', masking_key=masking_key)
        msg = i2b((0x81, 0x85, 0x37, 0xfa, 0x21, 0x3d, 0x7f, 0x9f,
                   0x4d, 0x51, 0x58))
        self.assertEqual(chunk, msg)

    def testParserBinary(self):
        s = self.parser()
        c = self.parser(kind=1)
        chunk = c.encode(self.bdata, opcode=2)
        frame = s.decode(chunk)
        self.assertTrue(frame)
        self.assertEqual(frame.body, self.bdata)
        #
        # Now try different masking key
        chunk = c.encode(self.large_bdata, opcode=2, masking_key=b'ciao')
        frame = s.decode(chunk)
        self.assertTrue(frame)
        self.assertEqual(frame.body, self.large_bdata)

    def testPartialParsing(self):
        s = self.parser()
        c = self.parser(kind=1)
        chunk = s.encode(self.large_bdata, opcode=2)
        #
        self.assertEqual(c.decode(chunk[:1]), None)
        self.assertEqual(c.decode(chunk[1:5]), None)
        self.assertEqual(c.decode(chunk[5:50]), None)
        frame = c.decode(chunk[50:])
        self.assertTrue(frame)
        self.assertEqual(frame.body, self.large_bdata)
        self.assertEqual(frame.opcode, 2)

    def test_multi_encode(self):
        s = self.parser()
        c = self.parser(kind=1)
        chunks = list(s.multi_encode(self.large_bdata, opcode=2,
                                     max_payload=6500))
        self.assertEqual(len(chunks), 11)
        #
        # Now decode them
        frames = []
        for chunk in chunks:
            frames.append(c.decode(chunk))
        for frame in frames[:-1]:
            self.assertFalse(frame.final)
        self.assertTrue(frames[-1].final)
        msg = b''.join((f.body for f in frames))
        self.assertEqual(msg, self.large_bdata)

    def test_bad_mask(self):
        s = self.parser()
        chunk = s.encode('hello')
        self.assertRaises(ProtocolError, s.decode, chunk)
        #
        # and the client
        c = self.parser(kind=1)
        chunk = c.encode('hello')
        self.assertRaises(ProtocolError, c.decode, chunk)

    def test_symmetric_mask(self):
        s = self.parser(kind=2)
        chunk = s.encode('Hello')
        self.assertEqual(s.decode(chunk).body, 'Hello')
        s = self.parser(kind=3)
        chunk = s.encode('Hello')
        self.assertEqual(s.decode(chunk).body, 'Hello')

    def test_parse_close(self):
        self.assertRaises(ProtocolError, parse_close, b'o')


@unittest.skipUnless(HAS_C_EXTENSIONS, "Requires C extensions")
class PyFrameTest(FrameTest):

    def parser(self, pyparser=True, **kw):
        return frame_parser(pyparser=True, **kw)

    def test_parsers(self):
        import pulsar.utils.websocket as ws
        self.assertNotEqual(ws.CFrameParser, ws.FrameParser)



import os

from pulsar.apps.http import SSLError, HttpClient

from tests.http import base


crt = os.path.join(os.path.dirname(__file__), 'ca_bundle')


class TestTlsHttpClient(base.TestHttpClient):
    with_tls = True

    async def test_verify(self):
        c = HttpClient()
        await self.wait.assertRaises(SSLError, c.get, self.httpbin())
        response = await c.get(self.httpbin(), verify=False)
        self.assertEqual(response.status_code, 200)
        response = await c.get(self.httpbin(), verify=crt)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.request.verify, crt)



import unittest

from pulsar.apps import wsgi


class TestHtml(unittest.TestCase):

    def test_html_factory(self):
        input = wsgi.html_factory('input', type='text')
        h = input(value='bla')
        self.assertTrue(h)
        self.assertEqual(h.attr('value'), 'bla')
        self.assertEqual(h.attr('type'), 'text')
        text = h.render()
        self.assertTrue(" type='text'" in text)
        self.assertTrue(" value='bla'" in text)

    def testEmpty(self):
        c = wsgi.Html('div')
        self.assertEqual(c._attr, None)
        self.assertEqual(c._classes, None)
        self.assertEqual(c._data, None)
        self.assertEqual(c._css, None)

    def test_html_repr(self):
        c = wsgi.Html('div', cn='bla', charset='utf-16')
        self.assertEqual(c.content_type, 'text/html; charset=utf-16')
        self.assertEqual(str(c), "<div class='bla'>")
        c = wsgi.Html(None, cn='bla')
        self.assertEqual(c.tag, None)
        self.assertEqual(str(c), "Html")

    def testClass(self):
        c = wsgi.Html('div').addClass('ciao').addClass('pippo')
        self.assertTrue(c.hasClass('ciao'))
        self.assertTrue(c.hasClass('pippo'))
        f = c.flatatt()
        self.assertTrue(f in (" class='ciao pippo'",
                              " class='pippo ciao'"))
        c = wsgi.Html('div').addClass('ciao pippo bla')
        self.assertTrue(c.hasClass('bla'))

    def testClassList(self):
        c = wsgi.Html('div', cn=['ciao', 'pippo', 'ciao'])
        self.assertEqual(len(c._classes), 2)
        self.assertTrue('ciao' in c._classes)
        self.assertTrue('pippo' in c._classes)
        self.assertTrue(c.hasClass('pippo'))

    def testRemoveClass(self):
        c = wsgi.Html('div', cn=['ciao', 'pippo', 'ciao', 'foo'])
        self.assertEqual(len(c._classes), 3)
        c.removeClass('sdjkcbhjsd smdhcbjscbsdcsd')
        self.assertEqual(len(c._classes), 3)
        c.removeClass('ciao sdjkbsjc')
        self.assertEqual(len(c._classes), 2)
        c.removeClass('pippo foo')
        self.assertEqual(len(c._classes), 0)

    def testData(self):
        c = wsgi.Html('div')
        c.data('food', 'pasta').data('city', 'Rome')
        self.assertEqual(c.data('food'), 'pasta')
        self.assertEqual(c.data('city'), 'Rome')
        f = c.flatatt()
        self.assertTrue(f in (" data-food='pasta' data-city='Rome'",
                              " data-city='Rome' data-food='pasta'"))
        c.data('food', 'risotto')
        self.assertEqual(c.data('food'), 'risotto')

    def testNestedData(self):
        random = ['bla', 3, 'foo']
        table = {'name': 'path',
                 'resizable': True}
        c = wsgi.Html('div')
        r = c.data({'table': table, 'random': random})
        self.assertEqual(r, c)
        self.assertEqual(c.data('table')['name'], 'path')
        self.assertEqual(c.data('table')['resizable'], True)
        self.assertEqual(c.data('random')[0], 'bla')
        attr = c.flatatt()
        self.assertTrue('data-table' in attr)
        c.data('table', {'resizable': False, 'rows': 40})
        self.assertEqual(c.data('table')['name'], 'path')
        self.assertEqual(c.data('table')['resizable'], False)
        self.assertEqual(c.data('table')['rows'], 40)

    def testEmptyAttr(self):
        c = wsgi.Html('input', type='text')
        c.attr('value', None)
        self.assertEqual(c.attr('value'), None)
        self.assertEqual(c.flatatt(), " type='text'")
        c.attr('value', '')
        self.assertEqual(c.attr('value'), '')
        self.assertTrue(" value=''" in c.flatatt())
        c.attr('value', 0)
        self.assertTrue(" value='0'" in c.flatatt())
        self.assertEqual(c.attr('value'), 0)

    def test_true_attribute(self):
        txt = wsgi.Html('div', bla=True).render()
        self.assertEqual(txt, '<div bla></div>')

    def test_option_empty_attribute(self):
        opt = wsgi.Html('option', '--------', value='')
        self.assertEqual(opt.attr('value'), '')
        text = opt.render()
        self.assertTrue(" value=''" in text)

    def test_textarea_value_attribute(self):
        opt = wsgi.Html('textarea', 'Hello World!')
        self.assertEqual(opt.attr('value'), None)
        self.assertEqual(opt.get_form_value(), 'Hello World!')
        text = opt.render()
        self.assertTrue("Hello World!" in text)

    def testHide(self):
        c = wsgi.Html('div', 'foo').hide()
        self.assertEqual(c.flatatt(), " style='display:none;'")
        c.show()
        self.assertEqual(c.flatatt(), "")


class TestWidgets(unittest.TestCase):

    def test_ancor(self):
        a = wsgi.Html('a', 'kaput', cn='bla', href='/abc/')
        self.assertEqual(a.attr('href'), '/abc/')
        ht = a.render()
        self.assertTrue('>kaput</a>' in ht)
        a = wsgi.Html('a', xxxx='ciao')
        self.assertTrue('xxxx' in a.attr())

    def testList(self):
        ul = wsgi.Html('ul')
        self.assertEqual(ul.tag, 'ul')
        self.assertEqual(len(ul.children), 0)
        ul = wsgi.Html('ul', 'a list item', 'another one')
        self.assertEqual(len(ul.children), 2)
        ht = ul.render()
        self.assertTrue('<ul>' in ht)
        self.assertTrue('</ul>' in ht)
        self.assertTrue('<li>a list item</li>' in ht)
        self.assertTrue('<li>another one</li>' in ht)

    def test_simple(self):
        html = wsgi.HtmlDocument()
        self.assertEqual(len(html.head.children), 5)
        self.assertEqual(len(html.body.children), 0)

    def testHead(self):
        html = wsgi.HtmlDocument()
        head = html.head
        self.assertTrue(head.meta)
        self.assertTrue(head.links)
        self.assertTrue(head.scripts)

    def test_meta(self):
        html = wsgi.HtmlDocument()
        meta = html.head.meta
        self.assertEqual(len(meta.children), 1)
        self.assertEqual(meta.tag, None)
        html.head.add_meta(name='bla')
        self.assertEqual(len(meta.children), 2)
        text = meta.render()
        self.assertEqual(text, "<meta charset='utf-8'>\n<meta name='bla'>\n")

    def test_document_empty_body(self):
        m = wsgi.HtmlDocument(title='test', bla='foo')
        self.assertTrue(m.head)
        self.assertTrue(m.body)
        self.assertEqual(m.head.title, 'test')
        txt = m.render()
        self.assertEqual(txt,
                         '\n'.join(('<!DOCTYPE html>',
                                    "<html bla='foo'>",
                                    '<head>',
                                    '<title>test</title>',
                                    "<meta charset='utf-8'>",
                                    '</head>',
                                    '<body>',
                                    '</body>',
                                    '</html>')))

    def test_document(self):
        m = wsgi.HtmlDocument(title='test')
        m.body.append(wsgi.Html('div', 'this is a test'))
        txt = m.render()
        self.assertEqual(txt,
                         '\n'.join(['<!DOCTYPE html>',
                                    '<html>',
                                    '<head>',
                                    '<title>test</title>',
                                    "<meta charset='utf-8'>",
                                    '</head>',
                                    '<body>',
                                    '<div>this is a test</div>'
                                    '</body>',
                                    '</html>']))




if __name__ == '__main__':
    import sys
    import pickle
    from multiprocessing import current_process
    from multiprocessing.spawn import import_main_path

    data = pickle.load(sys.stdin.buffer)
    current_process().authkey = data['authkey']
    sys.path = data['path']
    import_main_path(data['main'])
    impl = pickle.loads(data['impl'])

    from pulsar.async.concurrency import run_actor

    run_actor(impl)



from functools import partial
from collections import namedtuple
from copy import copy
from urllib.parse import urlparse, urljoin

from pulsar import OneTime, isawaitable, create_future, PulsarException
from pulsar.apps.ws import WebSocketProtocol, WS
from pulsar.utils.httpurl import REDIRECT_CODES, requote_uri, SimpleCookie
from pulsar.utils.websocket import SUPPORTED_VERSIONS, websocket_key


def noerror(callback):
    '''Decorator to run a callback of a :class:`.EventHandler`
    only when no errors occur
    '''
    def _(*response, **kw):
        if response[-1] and not kw.get('exc'):
            return callback(*response)

    return _


HTTP11 = 'HTTP/1.1'


def keep_alive(version, headers):
    """Check if to keep alive an HTTP connection.

    If the version is 1.1, we close the connection only if the ``connection``
    header is available and set to ``close``
    """
    if version == HTTP11:
        return not headers.has('connection', 'close')
    else:
        return headers.has('connection', 'keep-alive')


def response_content(resp, exc=None, **kw):
    b = resp.parser.recv_body()
    if b or resp._content is None:
        resp._content = resp._content + b if resp._content else b
    return resp._content


def _consumer(response, consumer):
    if response is not None:
        consumer = response
    return consumer


async def start_request(request, conn):
    response = conn.current_consumer()
    # bind request-specific events
    response.bind_events(**request.inp_params)
    if request.auth:
        response.bind_event('pre_request', request.auth)

    if request.stream:
        response.bind_event('data_processed', response.raw)
        response.start(request)
        await response.events['on_headers']
    else:
        response.bind_event('data_processed', response_content)
        response.start(request)
        await response.on_finished

    if hasattr(response.request_again, '__call__'):
        response = response.request_again(response)
        if isawaitable(response):
            response = await response

    return response


class request_again(namedtuple('request_again', 'method url params')):

    @property
    def status_code(self):
        return -1

    @property
    def headers(self):
        return ()


class TooManyRedirects(PulsarException):

    def __init__(self, response):
        self.response = response


class WebSocketClient(WebSocketProtocol):
    status_code = 101

    @property
    def _request(self):
        return self.handshake._request

    @property
    def headers(self):
        return self.handshake.headers

    def __getattr__(self, name):
        if not name.startswith('__'):
            return getattr(self.handshake, name)
        else:
            raise AttributeError("'%s' object has no attribute '%s'" %
                                 (self.__class__.__name__, name))


class Redirect:

    @noerror
    def __call__(self, response, exc=None):
        if (response.status_code in REDIRECT_CODES and
                'location' in response.headers and
                response.request.allow_redirects):
            response.request_again = self._do_redirect

    def _do_redirect(self, response):
        request = response.request
        # done with current response
        url = response.headers.get('location')
        # Handle redirection without scheme (see: RFC 1808 Section 4)
        if url.startswith('//'):
            parsed_rurl = urlparse(request.url)
            url = '%s:%s' % (parsed_rurl.scheme, url)
        # Facilitate non-RFC2616-compliant 'location' headers
        # (e.g. '/path/to/resource' instead of
        # 'http://domain.tld/path/to/resource')
        if not urlparse(url).netloc:
            url = urljoin(request.url,
                          # Compliant with RFC3986, we percent
                          # encode the url.
                          requote_uri(url))
        history = request.history
        if history and len(history) >= request.max_redirects:
            raise TooManyRedirects(response)

        params = request.inp_params.copy()
        params['history'] = copy(history) if history else []
        params['history'].append(response)
        if response.status_code == 303:
            method = 'GET'
            params.pop('data', None)
            params.pop('files', None)
        else:
            method = request.method
        response.request_again = request_again(method, url, params)
        return response


@noerror
def handle_cookies(response, exc=None):
    '''Handle response cookies.
    '''
    headers = response.headers
    request = response.request
    client = request.client
    response._cookies = c = SimpleCookie()
    if 'set-cookie' in headers or 'set-cookie2' in headers:
        for cookie in (headers.get('set-cookie2'),
                       headers.get('set-cookie')):
            if cookie:
                c.load(cookie)
        if client.store_cookies:
            client.cookies.extract_cookies(response, request)


class WebSocket:

    @property
    def websocket_key(self):
        if not hasattr(self, '_websocket_key'):
            self._websocket_key = websocket_key()
        return self._websocket_key

    @noerror
    def __call__(self, response, exc=None):
        request = response.request
        if request and urlparse(request.url).scheme in ('ws', 'wss'):
            headers = request.headers
            headers['connection'] = 'Upgrade'
            headers['upgrade'] = 'websocket'
            if 'Sec-WebSocket-Version' not in headers:
                headers['Sec-WebSocket-Version'] = str(max(SUPPORTED_VERSIONS))
            if 'Sec-WebSocket-Key' not in headers:
                headers['Sec-WebSocket-Key'] = self.websocket_key
            response.bind_event('on_headers', self.on_headers)

    @noerror
    def on_headers(self, response, exc=None):
        '''Websocket upgrade as ``on_headers`` event.'''

        if response.status_code == 101:
            connection = response.connection
            request = response.request
            handler = request.websocket_handler
            if not handler:
                handler = WS()
            parser = request.client.frame_parser(kind=1)
            consumer = partial(WebSocketClient, response, handler, parser)
            connection.upgrade(consumer)
            body = response.recv_body()
            response.finished()
            websocket = connection.current_consumer()
            websocket.data_received(body)
            response.request_again = lambda r: websocket


class Tunneling:
    '''A pre request callback for handling proxy tunneling.

    If Tunnelling is required, it writes the CONNECT headers and abort
    the writing of the actual request until headers from the proxy server
    are received.
    '''
    def __init__(self, loop):
        assert loop
        self._loop = loop

    @noerror
    def __call__(self, response, exc=None):
        # the pre_request handler
        request = response.request
        if request and request.tunnel:
            request.tunnel.apply(response, self)

    @noerror
    def on_headers(self, response, exc=None):
        '''Called back once the headers have arrived.'''
        if response.status_code == 200:
            response.request_again = self._tunnel_request
            response.finished()

    async def _tunnel_request(self, response):
        request = response.request.request
        connection = response.connection
        loop = connection._loop
        sock = connection.sock
        connection.transport.pause_reading()
        # await asyncio.sleep(0.01)
        # set a new connection_made event
        connection.events['connection_made'] = OneTime(loop=loop)
        connection._processed -= 1
        connection.producer._requests_processed -= 1
        #
        url = urlparse(request.url)
        await ssl_transport(loop, sock, connection, request._ssl, url.netloc)
        await connection.event('connection_made')
        response = await start_request(request, connection)
        return response


async def ssl_transport(loop, rawsock, connection, sslcontext, hostname):

    if hasattr(loop, '_make_legacy_ssl_transport'):
        # TODO: this is a hack because the create_connection does not work
        # with standard asyncio event loops
        waiter = create_future(loop)
        loop._make_legacy_ssl_transport(
            rawsock,
            connection,
            sslcontext,
            waiter,
            server_hostname=hostname
        )
        await waiter

    else:
        await loop.create_connection(
            lambda: connection,
            ssl=sslcontext,
            sock=rawsock,
            server_hostname=hostname
        )



# -*- coding: utf-8 -
"""Event driven concurrent framework for Python"""
import os

from .utils.version import get_version


VERSION = (1, 5, 0, 'final', 0)

__version__ = version = get_version(VERSION)


DEFAULT_PORT = 8060
ASYNC_TIMEOUT = None
SERVER_NAME = 'pulsar'
JAPANESE = b'\xe3\x83\x91\xe3\x83\xab\xe3\x82\xb5\xe3\x83\xbc'.decode('utf-8')
CHINESE = b'\xe8\x84\x89\xe5\x86\xb2\xe6\x98\x9f'.decode('utf-8')
HINDI = (b'\xe0\xa4\xaa\xe0\xa4\xb2\xe0\xa5\x8d'
         b'\xe0\xa4\xb8\xe0\xa4\xb0').decode('utf-8')
SERVER_SOFTWARE = "{0}/{1}".format(SERVER_NAME, version)

if os.environ.get('pulsar_speedup') == 'no':
    HAS_C_EXTENSIONS = False
else:
    HAS_C_EXTENSIONS = True
    try:
        from .utils import lib      # noqa
    except ImportError:
        HAS_C_EXTENSIONS = False

from .utils.exceptions import *     # noqa
from .utils import system           # noqa
platform = system.platform          # noqa
from .utils.config import *         # noqa
from .async import *                # noqa
from .apps import *                 # noqa
from .apps.data import data_stores  # noqa



import threading
from collections import deque

from greenlet import getcurrent

from pulsar import get_event_loop, create_future

from .utils import MustBeInChildGreenlet


class GreenLock:
    """A Locking primitive that is owned by a particular greenlet
    when locked.The main greenlet cannot acquire the lock.

    A primitive lock is in one of two states, 'locked' or 'unlocked'.

    It is created in the unlocked state. It has two basic methods,
    :meth:`.acquire` and :meth:`.release. When the state is unlocked,
    :meth:`.acquire` changes the state to locked and returns immediately.

    When the state is locked, :meth:`.acquire` blocks the current greenlet
    until a call to :meth:`.release` changes it to unlocked,
    then the :meth:`.acquire` call resets it to locked and returns.
    """
    def __init__(self, loop=None):
        self._loop = loop or get_event_loop()
        self._local = threading.local()
        self._local.locked = None
        self._queue = deque()

    def locked(self):
        """'Return the greenlet that acquire the lock or None.
        """
        return self._local.locked

    def acquire(self, timeout=None):
        """Acquires the lock if in the unlocked state otherwise switch
        back to the parent coroutine.
        """
        green = getcurrent()
        parent = green.parent
        if parent is None:
            raise MustBeInChildGreenlet('GreenLock.acquire in main greenlet')

        if self._local.locked:
            future = create_future(self._loop)
            self._queue.append(future)
            parent.switch(future)

        self._local.locked = green
        return self.locked()

    def release(self):
        """Release the lock.

        This method should only be called in the locked state;
        it changes the state to unlocked and returns immediately.
        If an attempt is made to release an unlocked lock,
        a RuntimeError will be raised.
        """
        if self._local.locked:
            while self._queue:
                future = self._queue.popleft()
                if not future.done():
                    return future.set_result(None)
            self._local.locked = None
        else:
            raise RuntimeError('release unlocked lock')

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, type, value, traceback):
        self.release()




cdef extern from "math.h":
    double log(double x)
    double sqrt(double x)


cdef double clog2 = log(2.)

cdef inline int int_max(int a, int b): return a if a >= b else b
cdef inline int int_min(int a, int b): return a if a >= b else b


# MSVC does not have log2!
cdef inline double Log2(double x):
    return log(x) / clog2



import shlex
from multiprocessing import current_process

import setuptools.command.test as orig


OPTION_MAP = {
    'io': 'event_loop'
}


def extend(values, extra):
    extra.extend(values)
    return extra


class Test(orig.test):
    test_suite = True
    start_coverage = False
    list_options = set([
        'log-level=',
        'test-plugins=',
        'test-modules=',
        'pulsar-args='
    ])
    user_options = [
        ('list-labels', 'l', 'List all test labels without performing tests'),
        ('coverage', None, 'Collect code coverage from all spawn actors'),
        ('coveralls', None, 'Publish coverage to coveralls'),
        ('sequential', None, 'Run test functions sequentially'),
        ('test-timeout=', None, 'Timeout for asynchronous tests'),
        ('log-level=', None, 'Logging level'),
        ('io=', None, 'Event Loop'),
        ('concurrency=', None, 'Concurrency'),
        ('test-plugins=', None, 'Test plugins'),
        ('test-modules=', None, 'Modules where to look for tests'),
        ('pulsar-args=', 'a',
         "Additional arguments to pass to pulsar test suite")]

    def initialize_options(self):
        for name, _, _ in self.user_options:
            setattr(self, self._slugify(name), None)

    def finalize_options(self):
        self.test_params = {}
        for name, _, _ in self.user_options:
            attr = self._slugify(name)
            value = getattr(self, attr)
            if value and name in self.list_options:
                value = shlex.split(value)
                setattr(self, attr, value)
            if value is not None:
                param_name = OPTION_MAP.get(attr, attr)
                self.test_params[param_name] = value
        self.test_args = self.pulsar_args or []

    def run_tests(self):
        if self.coverage and self.start_coverage:
            import coverage
            p = current_process()
            p._coverage = coverage.Coverage(data_suffix=True)
            coverage.process_startup()
            p._coverage.start()

        from pulsar.apps.test import TestSuite
        params = self.get_test_parameters()
        test_suite = TestSuite(argv=self.test_args,
                               **params)
        test_suite.start()

    def get_test_parameters(self):
        params = self.test_params
        params['verbosity'] = self.verbose+1
        return params

    def _slugify(self, name):
        return name.replace('-', '_').replace('=', '')


class Bench(Test):
    description = 'Run benchmarks with pulsar benchmark test suite'
    user_options = extend(Test.user_options, [
        ('repeat=', None, 'Number of repetitions'),
    ])

    def get_test_parameters(self):
        params = self.test_params
        params['benchmark'] = True
        params['verbosity'] = False
        if 'test_timeout' not in params:
            params['test_timeout'] = 300
        return params



from functools import wraps

from pulsar import isawaitable

from greenlet import greenlet, getcurrent


class MustBeInChildGreenlet(RuntimeError):
    """Raised when an operation must be performed in a child greenlet
    """


class GreenletWorker(greenlet):
    pass


def wait(value, must_be_child=False):
    '''Wait for a possible asynchronous value to complete.
    '''
    current = getcurrent()
    parent = current.parent
    if must_be_child and not parent:
        raise MustBeInChildGreenlet('Cannot wait on main greenlet')
    return parent.switch(value) if parent else value


def run_in_greenlet(callable):
    """Decorator to run a ``callable`` on a new greenlet.

    A ``callable`` decorated with this decorator returns a coroutine
    """
    @wraps(callable)
    async def _(*args, **kwargs):
        green = greenlet(callable)
        # switch to the new greenlet
        result = green.switch(*args, **kwargs)
        # back to the parent
        while isawaitable(result):
            # keep on switching back to the greenlet if we get a Future
            try:
                result = green.switch((await result))
            except Exception as exc:
                result = green.throw(exc)

        return green.switch(result)

    return _



import unittest

try:
    import requests
    http = requests.session()
except ImportError:
    http = None

from pulsar.apps.test import dont_run_with_thread


class TestRequest:

    async def setUp(self):
        http = self.client()
        response = await http.get(self.httpbin())
        self.assertEqual(str(response), '<Response [200]>')

    @dont_run_with_thread
    @unittest.skipUnless(http, "requires python requests library")
    def test_requests_get_200(self):
        response = http.get(self.httpbin(), verify=False,
                            proxies=self.proxies())
        self.assertEqual(str(response), '<Response [200]>')
        self.assertEqual(response.status_code, 200)
        self.assertTrue(response.content)
        self.assertEqual(response.url, self.httpbin())
        response = http.get(self.httpbin('get'), verify=False,
                            proxies=self.proxies())
        self.assertEqual(response.status_code, 200)



'''Tests for arbiter and monitors.'''
import unittest
import asyncio

import pulsar
from pulsar import send, spawn, ACTOR_ACTION_TIMEOUT
from pulsar.apps.test import ActorTestMixin, test_timeout

from tests.async import cause_timeout, cause_terminate, wait_for_stop


class TestArbiterProcess(ActorTestMixin, unittest.TestCase):
    concurrency = 'process'

    def test_arbiter_object(self):
        '''Test the arbiter in its process domain'''
        arbiter = pulsar.get_actor()
        self.assertEqual(arbiter, pulsar.arbiter())
        self.assertTrue(arbiter.is_arbiter())
        self.assertEqual(arbiter.impl.kind, 'arbiter')
        self.assertEqual(arbiter.aid, 'arbiter')
        self.assertEqual(arbiter.name, 'arbiter')
        self.assertTrue(arbiter.monitors)
        self.assertEqual(arbiter.exit_code, None)
        info = arbiter.info()
        self.assertTrue('server' in info)
        server = info['server']
        self.assertEqual(server['state'], 'running')

    def test_arbiter_mailbox(self):
        arbiter = pulsar.get_actor()
        mailbox = arbiter.mailbox
        self.assertFalse(hasattr(mailbox, 'request'))
        # Same for all monitors mailboxes
        for monitor in arbiter.monitors.values():
            mailbox = monitor.mailbox
            self.assertFalse(hasattr(mailbox, 'request'))

    def test_registered(self):
        '''Test the arbiter in its process domain'''
        arbiter = pulsar.get_actor()
        self.assertTrue(arbiter.is_arbiter())
        self.assertTrue(arbiter.registered)
        self.assertTrue('arbiter' in arbiter.registered)
        self.assertTrue('test' in arbiter.registered)

    @test_timeout(2*ACTOR_ACTION_TIMEOUT)
    async def test_spawning_in_arbiter(self):
        arbiter = pulsar.get_actor()
        self.assertEqual(arbiter.name, 'arbiter')
        self.assertTrue(len(arbiter.monitors) >= 1)
        name = 'testSpawning-%s' % self.concurrency
        future = spawn(name=name, concurrency=self.concurrency)
        self.assertTrue(future.aid in arbiter.managed_actors)
        proxy = await future
        self.assertEqual(future.aid, proxy.aid)
        self.assertEqual(proxy.name, name)
        self.assertTrue(proxy.aid in arbiter.managed_actors)
        self.assertEqual(proxy, arbiter.get_actor(proxy.aid))
        #
        await asyncio.sleep(1)
        # Stop the actor
        result = await send(proxy, 'stop')
        self.assertEqual(result, None)
        #
        result = await wait_for_stop(self, proxy.aid)
        self.assertEqual(result, None)

    def test_bad_monitor(self):
        arbiter = pulsar.get_actor()
        self.assertTrue(arbiter.monitors)
        name = list(arbiter.monitors.values())[0].name
        self.assertRaises(KeyError, arbiter.add_monitor, name)

    @test_timeout(2*ACTOR_ACTION_TIMEOUT)
    async def test_terminate(self):
        arbiter = pulsar.get_actor()
        self.assertTrue(arbiter.is_arbiter())
        name = 'bogus-term-%s' % self.concurrency
        proxy = await self.spawn_actor(name=name, timeout=1)
        self.assertEqual(proxy.name, name)
        self.assertTrue(proxy.aid in arbiter.managed_actors)
        proxy = arbiter.managed_actors[proxy.aid]
        #
        await send(proxy, 'run', cause_terminate)
        #
        # The arbiter should now terminate the actor
        await wait_for_stop(self, proxy.aid, True)
        #
        self.assertTrue(proxy.stopping_start)

    async def test_actor_timeout(self):
        """Test a bogus actor for timeout"""
        arbiter = pulsar.get_actor()
        self.assertTrue(arbiter.is_arbiter())
        name = 'bogus-timeout-%s' % self.concurrency
        proxy = await self.spawn_actor(name=name, timeout=1)
        self.assertEqual(proxy.name, name)
        self.assertTrue(proxy.aid in arbiter.managed_actors)
        proxy = arbiter.managed_actors[proxy.aid]
        await send(proxy, 'run', cause_timeout)
        # The arbiter should soon start to stop the actor
        await wait_for_stop(self, proxy.aid, True)
        #
        self.assertTrue(proxy.stopping_start)
        self.assertFalse(proxy.aid in arbiter.managed_actors)



'''
A list of all Exception specific to pulsar library.
'''
import traceback

from .httpurl import Headers


__all__ = ['PulsarException',
           'MonitorStarted',
           'ImproperlyConfigured',
           'CommandError',
           'CommandNotFound',
           'ProtocolError',
           'EventAlreadyRegistered',
           'InvalidOperation',
           'HaltServer',
           'LockError',
           #
           # HTTP client exception
           'HttpRequestException',
           'HttpConnectionError',
           'HttpProxyError',
           'SSLError',
           #
           # HTTP Exceptions
           'HttpException',
           'HttpRedirect',
           'BadRequest',
           'Http401',
           'PermissionDenied',
           'Http404',
           'MethodNotAllowed',
           'HttpGone',
           'Unsupported',
           'UnprocessableEntity',
           #
           'format_traceback']


class PulsarException(Exception):
    '''Base class of all Pulsar exceptions.'''


class MonitorStarted(PulsarException):
    exit_code = 0


class ImproperlyConfigured(PulsarException):
    '''A :class:`PulsarException` raised when an inconsistent configuration
    has occurred.

    .. attribute:: exit_code

        the exit code when rising this exception is set to 2. This will
        cause pulsar to log the error rather than the full stack trace.
    '''
    exit_code = 2


class CommandError(PulsarException):
    pass


class CommandNotFound(CommandError):

    def __init__(self, name):
        super().__init__('Command "%s" not available' % name)


class ProtocolError(PulsarException):
    '''A :class:`PulsarException` raised when the protocol encounter
    unexpected data.

    It will close the socket connection
    '''
    status_code = None

    def ProtocolError(self, msg=None, status_code=None):
        super().__init__(msg)
        self.status_code = status_code


class EventAlreadyRegistered(PulsarException):
    pass


class InvalidOperation(PulsarException):
    '''An invalid operation in pulsar'''
    pass


class HaltServer(BaseException):
    ''':class:`BaseException` raised to stop a running server.

    When ``exit_code`` is greater than 1, it is considered an expected
    failure and therefore the full stack trace is not logged.'''
    def __init__(self, reason='Exiting server.', exit_code=3):
        super().__init__(reason)
        self.exit_code = exit_code


class LockError(RuntimeError):
    pass


# #################################################################### HTTP
class HttpRequestException(IOError):
    "Base for all HTTP related errors"
    def __init__(self, *args, **kwargs):
        """
        Initialize RequestException with `request` and `response` objects.
        """
        response = kwargs.pop('response', None)
        self.response = response
        self.request = kwargs.pop('request', None)
        if (response is not None and not self.request and
                hasattr(response, 'request')):
            self.request = self.response.request
        super().__init__(*args, **kwargs)


class HttpConnectionError(HttpRequestException):
    """A Connection error occurred."""


class HttpProxyError(HttpConnectionError):
    """A proxy error occurred."""


class SSLError(HttpConnectionError):
    """An SSL error occurred."""


# #################################################################### HTTP
http_errors = {}


def httperror(c):
    http_errors[c.status] = c
    return c


@httperror
class HttpException(PulsarException):
    '''The base class of all ``HTTP`` server exceptions

    Introduces the following attributes:

    .. attribute:: status

        The numeric status code for the exception (ex 500 for server error).

        Default: ``500``.

    .. attribute:: headers

        Additional headers to add to the client response.
    '''
    status = 500

    def __init__(self, msg='', status=None, handler=None, strict=False,
                 headers=None, content_type=None):
        self.status = status or self.status
        self.code = self.status  # for compatibility with HTTPError
        self.handler = handler
        self.strict = strict
        self.content_type = content_type
        self._headers = Headers.make(headers)
        super().__init__(msg)

    @property
    def headers(self):
        return list(self._headers)


@httperror
class HttpRedirect(HttpException):
    '''An :class:`HttpException` for redirects.

    The :attr:`~HttpException.status` is set to ``302`` by default.
    '''
    status = 302

    def __init__(self, location, status=None, headers=None, **kw):
        headers = Headers.make(headers)
        headers['location'] = location
        super().__init__(status=status or self.status, headers=headers, **kw)

    @property
    def location(self):
        '''The value in the ``Location`` header entry.
        '''
        return self._headers['location']


@httperror
class BadRequest(HttpException):
    status = 400


@httperror
class Http401(HttpException):
    status = 401

    def __init__(self, auth, msg=''):
        headers = [('WWW-Authenticate', auth)]
        super().__init__(msg=msg, headers=headers)


@httperror
class PermissionDenied(HttpException):
    '''An :class:`HttpException` with default ``403`` status code.'''
    status = 403


@httperror
class Http404(HttpException):
    '''An :class:`HttpException` with default ``404`` status code.'''
    status = 404


@httperror
class MethodNotAllowed(HttpException):
    '''An :class:`HttpException` with default ``405`` status code.'''
    status = 405


@httperror
class HttpGone(HttpException):
    status = 410


@httperror
class Unsupported(HttpException):
    status = 415


@httperror
class UnprocessableEntity(HttpException):
    status = 422


def format_traceback(exc):
    return traceback.format_exception(exc.__class__, exc, exc.__traceback__)



import os
import threading
import logging
import asyncio
from concurrent.futures import ThreadPoolExecutor

from collections import OrderedDict
from threading import current_thread

from asyncio import Future

from pulsar.utils.config import Global
from pulsar.utils.system import current_process

from asyncio import ensure_future
from inspect import isawaitable


__all__ = ['get_event_loop',
           'new_event_loop',
           'get_actor',
           'cfg',
           'cfg_value',
           'isfuture',
           'create_future',
           'is_mainthread',
           'process_data',
           'thread_data',
           'logger',
           'NOTHING',
           'EVENT_LOOPS',
           'Future',
           'reraise',
           'isawaitable',
           'ensure_future',
           'CANCELLED_ERRORS']


_EVENT_LOOP_CLASSES = (asyncio.AbstractEventLoop,)
CANCELLED_ERRORS = (asyncio.CancelledError,)


def reraise(tp, value, tb=None):
    if value.__traceback__ is not tb:
        raise value.with_traceback(tb)
    raise value


def isfuture(x):
    return isinstance(x, Future)


def create_future(loop=None):
    loop = loop or get_event_loop()
    try:
        return loop.create_future()
    except AttributeError:
        return asyncio.Future(loop=loop)


LOGGER = logging.getLogger('pulsar')
NOTHING = object()
EVENT_LOOPS = OrderedDict()

DefaultLoopClass = asyncio.get_event_loop_policy()._loop_factory


def make_loop_factory(selector):

    def loop_factory():
        return DefaultLoopClass(selector())

    return loop_factory


for selector in ('Epoll', 'Kqueue', 'Poll', 'Select'):
    name = '%sSelector' % selector
    selector_class = getattr(asyncio.selectors, name, None)
    if selector_class:
        EVENT_LOOPS[selector.lower()] = make_loop_factory(selector_class)


try:    # add uvloop if available
    import uvloop
    EVENT_LOOPS['uv'] = uvloop.Loop
except Exception:     # pragma    nocover
    pass


if os.environ.get('BUILDING-PULSAR-DOCS') == 'yes':     # pragma nocover
    default_loop = (
        'uvloop if available, epoll on linux, '
        'kqueue on mac, select on windows'
    )
elif EVENT_LOOPS:
    default_loop = tuple(EVENT_LOOPS)[0]
else:
    default_loop = None


if default_loop:
    class EventLoopSetting(Global):
        name = "event_loop"
        flags = ["--io"]
        choices = tuple(EVENT_LOOPS)
        default = default_loop
        desc = """\
            Specify the event loop used for I/O event polling.

            The default value is the best possible for the system running the
            application.
            """

get_event_loop = asyncio.get_event_loop
new_event_loop = asyncio.new_event_loop


def is_mainthread(thread=None):
    '''Check if thread is the main thread.

    If ``thread`` is not supplied check the current thread
    '''
    thread = thread if thread is not None else current_thread()
    return isinstance(thread, threading._MainThread)


def logger(loop=None, logger=None):
    return getattr(loop or get_event_loop(), 'logger', LOGGER)


def process_data(name=None):
    '''Fetch the current process local data dictionary.

    If ``name`` is not ``None`` it returns the value at``name``,
    otherwise it return the process data dictionary
    '''
    ct = current_process()
    if not hasattr(ct, '_pulsar_local'):
        ct._pulsar_local = {}
    loc = ct._pulsar_local
    return loc.get(name) if name else loc


def thread_data(name, value=NOTHING, ct=None):
    '''Set or retrieve an attribute ``name`` from thread ``ct``.

    If ``ct`` is not given used the current thread. If ``value``
    is None, it will get the value otherwise it will set the value.
    '''
    ct = ct or current_thread()
    if is_mainthread(ct):
        loc = process_data()
    elif not hasattr(ct, '_pulsar_local'):
        ct._pulsar_local = loc = {}
    else:
        loc = ct._pulsar_local
    if value is not NOTHING:
        if name in loc:
            if loc[name] is not value:
                raise RuntimeError(
                    '%s is already available on this thread' % name)
        else:
            loc[name] = value
    return loc.get(name)


def get_actor():
    return thread_data('actor')


def set_actor(actor):
    return thread_data('actor', actor)


def cfg():
    actor = get_actor()
    if actor:
        return actor.cfg


def cfg_value(setting, value=None):
    if value is None:
        actor = get_actor()
        if actor:
            return actor.cfg.get(setting)
    return value


class EventLoopPolicy(asyncio.DefaultEventLoopPolicy):

    def __init__(self, name, workers, debug):
        super().__init__()
        self.name = name
        self.workers = workers
        self.debug = debug

    @property
    def _local(self):
        l = getattr(current_process(), '_event_loop_policy', None)
        if l is None:
            self._local = l = self._Local()
        return l

    @_local.setter
    def _local(self, v):
        current_process()._event_loop_policy = v

    def _loop_factory(self):
        loop = EVENT_LOOPS[self.name]()
        loop.set_default_executor(ThreadPoolExecutor(self.workers))
        if self.debug:
            loop.set_debug(True)
        return loop



import asyncio


class StreamConsumedError(Exception):
    """Same name as request error"""
    pass


class HttpStream:
    """An asynchronous streaming body for an HTTP response
    """
    def __init__(self, response):
        self._response = response
        self._streamed = False
        self._queue = asyncio.Queue()

    def __repr__(self):
        return repr(self._response)
    __str__ = __repr__

    @property
    def done(self):
        """Check if the stream is finished
        """
        return self._response.on_finished.fired()

    async def read(self, n=None):
        """Read all content
        """
        if self._streamed:
            return b''
        buffer = []
        async for body in self:
            buffer.append(body)
        return b''.join(buffer)

    def close(self):
        pass

    def __iter__(self):
        return _start_iter(self)

    def __next__(self):
        if self.done:
            try:
                return self._queue.get_nowait()
            except asyncio.QueueEmpty:
                raise StopIteration
        else:
            return self._queue.get()

    async def __aiter__(self):
        return _start_iter(self)

    async def __anext__(self):
        if self.done:
            try:
                return self._queue.get_nowait()
            except asyncio.QueueEmpty:
                raise StopAsyncIteration
        else:
            return await self._queue.get()

    def __call__(self, response, exc=None, **kw):
        if self._streamed and response.parser.is_headers_complete():
            assert response is self._response
            self._queue.put_nowait(response.recv_body())


def _start_iter(self):
    if self._streamed:
        raise StreamConsumedError
    self._streamed = True
    self(self._response)
    return self



import signal
from time import time

import pulsar
from pulsar import send, spawn, get_application


def add(actor, a, b):
    return actor.name, a+b


def wait_for_stop(test, aid, terminating=False):
    '''Wait for an actor to stop'''
    arbiter = pulsar.arbiter()
    waiter = pulsar.Future(loop=arbiter._loop)

    def remove():
        test.assertEqual(arbiter.remove_callback('periodic_task', check), 1)
        waiter.set_result(None)

    def check(caller, **kw):
        test.assertEqual(caller, arbiter)
        if not terminating:
            test.assertFalse(aid in arbiter.managed_actors)
        elif aid in arbiter.managed_actors:
            return
        arbiter._loop.call_soon(remove)

    arbiter.bind_event('periodic_task', check)
    return waiter


async def get_test(_):
    app = await get_application('test')
    return app.cfg


def check_environ(actor, name):
    import os
    return os.environ.get(name)


async def spawn_actor_from_actor(actor, name):
    actor2 = await spawn(name=name)
    pong = await send(actor2, 'ping')
    assert pong == 'pong', 'no pong from actor'
    t1 = time()
    # cover the notify from a fron actor
    t2 = await send(actor2, 'notify', {})
    assert t2 >= t1

    return actor2.aid


def cause_timeout(actor):
    if actor.next_periodic_task:
        actor.next_periodic_task.cancel()
    else:
        actor.event_loop.call_soon(cause_timeout, actor)


def cause_terminate(actor):
    if actor.next_periodic_task:
        actor.next_periodic_task.cancel()
        # hijack the SIGTERM
        actor.impl.kill = kill_hack(actor.impl.kill)
        actor.stop = lambda exc=None, exit_code=None: False
    else:
        actor._loop.call_soon(cause_timeout, actor)


def kill_hack(kill):

    def _(sig):
        if sig == signal.SIGKILL:
            kill(sig)

    return _


def close_mailbox(actor, close=False):
    if not close:
        actor._loop.call_later(0.5, close_mailbox, actor, True)
        return True
    # just for coverage
    assert repr(actor.mailbox)
    # close mailbox
    actor.mailbox.close()



'''Test cases for code not covered in standard test cases'''
import tempfile
import shutil
import unittest


class PulsarCoverage(unittest.TestCase):

    def test_profile_plugin(self):
        from pulsar.apps.test.plugins import profile
        p = profile.Profile()
        p.config.set('profile', True)
        p.configure(p.config)
        p.profile_temp_path = tempfile.mkdtemp()
        self.assertFalse(p.on_end())
        shutil.rmtree(p.profile_temp_path)
        #
        lines = list(profile.data_stream(['', 'a b c d e f']))
        self.assertEqual(lines, [])



'''tests the httpurl stand-alone script.'''
import time
import unittest

from pulsar.utils.httpurl import (Headers, CacheControl,
                                  urlquote, unquote_unreserved, requote_uri,
                                  remove_double_slash, appendslash, capfirst,
                                  encode_multipart_formdata, http_date,
                                  cookiejar_from_dict)
from pulsar.apps.http import Auth, HTTPBasicAuth, HTTPDigestAuth


class TestAuth(unittest.TestCase):

    def test_auth(self):
        auth = Auth()
        self.assertRaises(NotImplementedError, auth, None)
        self.assertEqual(str(auth), repr(auth))

    def test_basic_auth(self):
        auth = HTTPBasicAuth('bla', 'foo')
        self.assertEqual(str(auth), 'Basic: bla')

    def test_digest_auth(self):
        auth = HTTPDigestAuth('bla', options={'realm': 'fake realm'})
        self.assertEqual(auth.type, 'digest')
        self.assertEqual(auth.username, 'bla')
        self.assertEqual(auth.password, None)
        self.assertEqual(auth.options['realm'], 'fake realm')

    def test_CacheControl(self):
        headers = Headers()
        c = CacheControl()
        self.assertFalse(c.private)
        self.assertFalse(c.maxage)
        c(headers)
        self.assertEqual(headers['cache-control'], 'no-cache')
        c = CacheControl(maxage=3600)
        c(headers)
        self.assertEqual(headers['cache-control'], 'max-age=3600, public')
        c = CacheControl(maxage=3600, private=True)
        c(headers)
        self.assertEqual(headers['cache-control'], 'max-age=3600, private')
        c = CacheControl(maxage=3600, must_revalidate=True)
        c(headers)
        self.assertEqual(headers['cache-control'],
                         'max-age=3600, public, must-revalidate')
        c = CacheControl(maxage=3600, proxy_revalidate=True)
        c(headers)
        self.assertEqual(headers['cache-control'],
                         'max-age=3600, public, proxy-revalidate')
        c = CacheControl(maxage=3600, proxy_revalidate=True,
                         nostore=True)
        c(headers)
        self.assertEqual(headers['cache-control'],
                         'no-store, no-cache, must-revalidate, max-age=0')


class TestTools(unittest.TestCase):

    def test_quote_unreserved(self):
        '''Test a string of unreserved characters'''
        s = 'a~b_(c-d).'
        qs = urlquote(s)
        self.assertTrue('%' in qs)
        uqs = unquote_unreserved(qs)
        self.assertEqual(uqs, s)
        self.assertEqual(requote_uri(s), s)
        self.assertEqual(requote_uri(qs), s)

    def test_quote_unreserved_percent(self):
        s = 'a=3.5%;b=2%'
        qs = urlquote(s)
        self.assertTrue('%' in qs)
        uqs = unquote_unreserved(qs)
        self.assertNotEqual(uqs, s)
        s = 'a~b_(c-d).'
        qs = urlquote(s) + '%f'
        uqs = unquote_unreserved(qs)
        self.assertEqual(uqs, s+'%f')

    def test_remove_double_slash(self):
        r = remove_double_slash
        self.assertEqual(r('/bla//foo/'), '/bla/foo/')
        self.assertEqual(r('/bla/////////foo//////////'), '/bla/foo/')
        self.assertEqual(r('/bla/foo/'), '/bla/foo/')

    def test_appendslash(self):
        self.assertEqual(appendslash('bla'), 'bla/')
        self.assertEqual(appendslash('bla/'), 'bla/')

    def test_capfirst(self):
        c = capfirst
        self.assertEqual(c('blA'), 'Bla')
        self.assertEqual(c(''), '')
        self.assertEqual(c('bOlA'), 'Bola')

    def test_encode_multipart_formdata(self):
        data, ct = encode_multipart_formdata([('bla', 'foo'),
                                              ('foo', ('pippo', 'pluto'))])
        idx = data.find(b'\r\n')
        boundary = data[2:idx].decode('utf-8')
        self.assertEqual(ct, 'multipart/form-data; boundary=%s' % boundary)

    def test_http_date(self):
        now = time.time()
        fmt = http_date(now)
        self.assertTrue(fmt.endswith(' GMT'))
        self.assertEqual(fmt[3:5], ', ')

    def test_cookiejar_from_dict(self):
        j = cookiejar_from_dict({'bla': 'foo'}, None)
        self.assertEqual(len(j), 1)
        self.assertEqual(j, cookiejar_from_dict(None, j))
        j2 = cookiejar_from_dict({'pippo': 'pluto'}, None, j)
        self.assertEqual(len(j2), 2)



from random import randint
import unittest

from pulsar.utils.structures import Skiplist
from pulsar.apps.test import populate


class TestSkiplist(unittest.TestCase):
    skiplist = Skiplist

    def random(self, size=100, score_min=-10, score_max=10):
        scores = populate('float', size, score_min, score_max)
        data = populate('string', size, min_len=5, max_len=10)
        return self.skiplist(zip(scores, data))

    def test_extend(self):
        sl = self.skiplist()
        sl.extend([(94, 'bla'), (-5, 'foo')])
        self.assertEqual(len(sl), 2)

    def test_count(self):
        sl = self.skiplist()
        self.assertEqual(sl.count(-2, 2), 0)
        sl.insert(1, 'bla')
        self.assertEqual(sl.count(-2, 2), 1)
        sl = self.random()
        self.assertEqual(sl.count(float('-inf'), float('+inf')), 100)
        n = 0
        for score, _ in sl:
            n += int(score >= -5 and score <= 5)
        self.assertEqual(sl.count(-5, 5), n)
        #
        # test exclusive
        sl = self.skiplist(((1, 'a1'), (2, 'a2'), (3, 'a3')))
        self.assertEqual(sl.count(1, 2), 2)
        self.assertEqual(sl.count(1, 2, include_min=False), 1)
        self.assertEqual(sl.count(1, 2, include_max=False), 1)
        self.assertEqual(sl.count(1, 2, include_min=False,
                                  include_max=False), 0)

    def test_range_by_score(self):
        sl = self.skiplist()
        self.assertEqual(tuple(sl.range_by_score(float('-inf'), 5)), ())
        sl.insert(1, 'bla')
        self.assertEqual(tuple(sl.range_by_score(-3, 5)), ('bla',))
        sl.insert(1.5, 'foo')
        sl.insert(0.3, 'pippo')
        self.assertEqual(tuple(sl.range_by_score(0.9, 1.1)), ('bla',))
        self.assertEqual(tuple(sl.range_by_score(1, 2)), ('bla', 'foo'))
        self.assertEqual(tuple(sl.range_by_score(1, 2, include_min=False)),
                         ('foo',))
        self.assertEqual(tuple(sl.range_by_score(1, 1.5)), ('bla', 'foo'))
        self.assertEqual(tuple(sl.range_by_score(1, 1.5, include_max=False)),
                         ('bla',))
        self.assertEqual(tuple(sl.range_by_score(-1, 2, start=1)),
                         ('bla', 'foo'))
        self.assertEqual(tuple(sl.range_by_score(-1, 2, start=1, num=1)),
                         ('bla',))

    def test_remove_range(self):
        sl = self.skiplist()
        self.assertEqual(sl.remove_range(0, 3), 0)
        sl.insert(1, 'bla')
        self.assertEqual(sl.remove_range(0, 3), 1)
        sl = self.random(10)
        li = list(sl)
        lir = li[:4] + li[7:]
        self.assertEqual(sl.remove_range(4, 7), 3)
        li2 = list(sl)
        self.assertEqual(li2, lir)
        #
        sl = self.random()
        li = list(sl)
        lir = li[:5] + li[-8:]
        self.assertEqual(sl.remove_range(5, -8), len(li[5:-8]))
        li2 = list(sl)
        self.assertEqual(li2, lir)
        #
        sl = self.random()
        c = 0
        while sl:
            c += 1
            index = randint(0, len(sl)-1)
            self.assertEqual(sl.remove_range(index, index+1), 1)
        self.assertEqual(c, 100)

    def test_remove_range_by_score(self):
        sl = self.skiplist()
        self.assertEqual(sl.remove_range_by_score(0, 3), 0)
        sl.insert(1, 'bla')
        self.assertEqual(sl.remove_range_by_score(0, 3), 1)



import os
import sys

import coverage
from coverage.report import Reporter

from pulsar import new_event_loop
from pulsar.apps.http import HttpClient
from pulsar.utils.system import json
from pulsar.utils.version import gitrepo


COVERALLS_URL = 'https://coveralls.io/api/v1/jobs'


class CoverallsReporter(Reporter):

    def report(self, morfs):
        self.ret = []
        self.report_files(self._coverall, morfs)
        return self.ret

    def _coverall(self, fr, analysis):
        try:
            with open(fr.filename) as fp:
                source = fp.readlines()
        except IOError:
            if not self.config.ignore_errors:
                raise

        if self.config.strip_dirs:
            filename = fr.filename
            for dir in self.config.strip_dirs:
                if filename.startswith(dir):
                    filename = filename.replace(dir, '').lstrip('/')
                    break
        else:
            filename = fr.relname

        self.ret.append({
            'name': filename,
            'source': ''.join(source).rstrip(),
            'coverage': list(self._coverage_list(source, analysis)),
        })

    def _coverage_list(self, source, analysis):
        for lineno, line in enumerate(source, 1):
            if lineno in analysis.statements:
                yield int(lineno not in analysis.missing)
            else:
                yield None


class Coverage(coverage.Coverage):

    def coveralls(self, morfs=None, strip_dirs=None, **kw):
        self.config.from_args(**kw)
        self.config.strip_dirs = strip_dirs
        reporter = CoverallsReporter(self, self.config)
        return reporter.report(morfs)


def coveralls(http=None, url=None,
              data_file=None, repo_token=None,
              git=None, service_name=None,
              service_job_id=None, strip_dirs=None,
              ignore_errors=False, stream=None):
    '''Send a coverage report to coveralls.io.

    :param http: optional http client
    :param url: optional url to send data to. It defaults to ``coveralls``
        api url.
    :param data_file: optional data file to load coverage data from. By
        default, coverage uses ``.coverage``.
    :param repo_token: required when not submitting from travis.

    https://coveralls.io/docs/api
    '''
    if repo_token is None and os.path.isfile('.coveralls-repo-token'):
        with open('.coveralls-repo-token') as f:
            repo_token = f.read().strip()

    if strip_dirs is None:
        strip_dirs = [os.getcwd()]

    stream = stream or sys.stdout
    coverage = Coverage(data_file=data_file)
    coverage.load()
    if http is None:
        http = HttpClient(loop=new_event_loop())

    if not git:
        try:
            git = gitrepo()
        except Exception:   # pragma    nocover
            pass

    data = {'source_files': coverage.coveralls(strip_dirs=strip_dirs,
                                               ignore_errors=ignore_errors)}

    if git:
        data['git'] = git

    if os.environ.get('TRAVIS'):
        data['service_name'] = service_name or 'travis-ci'
        data['service_job_id'] = os.environ.get('TRAVIS_JOB_ID')
    else:
        assert repo_token, 'Requires repo_token if not submitting from travis'

    if repo_token:
        data['repo_token'] = repo_token
    url = url or COVERALLS_URL
    stream.write('Submitting coverage report to %s\n' % url)
    response = http.post(url, files={'json_file': json.dumps(data)})
    stream.write('Response code: %s\n' % response.status_code)
    try:
        info = response.json()
        code = 0
        if 'error' in info:
            stream.write('An error occurred while sending coverage'
                         ' report to coverall.io')
            code = 1
        stream.write('\n%s\n' % info['message'])
    except Exception:
        code = 1
        stream.write('Critical error %s\n' % response.status_code)
    return code



import signal

from pulsar.utils import autoreload
from pulsar import system


class ProcessMixin:

    def is_process(self):
        return True

    def before_start(self, actor):  # pragma    nocover
        actor.start_coverage()
        self._install_signals(actor)
        if actor.cfg.reload and self.is_arbiter():
            autoreload.start()

    def handle_int(self, actor, sig):
        actor.logger.warning("got %s - stopping", system.SIG_NAMES.get(sig))
        self.stop(actor, exit_code=int(sig))

    handle_term = handle_int
    handle_quit = handle_int
    handle_abrt = handle_int

    def handle_winch(self, actor, sig):
        actor.logger.debug("ignore %s", system.SIG_NAMES.get(sig))

    def _install_signals(self, actor):
        proc_name = actor.cfg.proc_name
        if proc_name:
            if not self.is_arbiter():
                name = actor.name.split('.')[0]
                proc_name = "%s-%s" % (proc_name, name)
            if system.set_proctitle(proc_name):
                actor.logger.debug('Set process title to %s',
                                   system.get_proctitle())
        system.set_owner_process(actor.cfg.uid, actor.cfg.gid)
        actor.logger.debug('Installing signals')
        loop = actor._loop
        for sig in system.SIGNALS:
            name = system.SIG_NAMES.get(sig)
            if name:
                handler = getattr(self, 'handle_%s' % name.lower(), None)
                if handler:
                    loop.add_signal_handler(sig, handler, actor, sig)

    def _remove_signals(self, actor):
        actor.logger.debug('Remove signal handlers')
        for sig in system.SIGNALS:
            try:
                actor._loop.remove_signal_handler(sig)
            except Exception:
                pass
        if actor._loop.is_running():  # pragma nocover
            actor.logger.critical('Event loop still running when stopping')
            actor._loop.stop()
        else:
            actor.logger.debug('Close event loop')
            actor._loop.close()


def signal_from_exitcode(sig):
    if sig in QUIT:
        sig = signal.SIGQUIT
    return sig


QUIT = set((signal.SIGINT, signal.SIGQUIT))



'''An asynchronous multi-process `HTTP proxy server`_. It works for both
``http`` and ``https`` (tunneled) requests.


Managing Headers
=====================
It is possible to add middleware to manipulate the original request headers.
If the header middleware is
an empty list, the proxy passes requests and responses unmodified.
This is an implementation for a forward-proxy which can be used
to retrieve any type of source from the Internet.

To run the server::

    python manage.py

An header middleware is a callable which receives the wsgi *environ* and
the list of request *headers*. By default the example uses:

.. autofunction:: x_forwarded_for

To run with different headers middleware create a new script and do::

    from proxyserver.manage import server

    if __name__ == '__main__':
        server(headers_middleware=[...]).start()

Implementation
===========================

.. autoclass:: ProxyServerWsgiHandler
   :members:
   :member-order:


.. _`HTTP proxy server`: http://en.wikipedia.org/wiki/Proxy_server
'''
import logging
from functools import partial

import pulsar
from pulsar import HttpException, ensure_future, create_future
from pulsar.apps import wsgi, http
from pulsar.apps.http.plugins import noerror
from pulsar.utils.httpurl import Headers, ENCODE_BODY_METHODS
from pulsar.utils.log import LocalMixin, local_property


SERVER_SOFTWARE = 'Pulsar-proxy-server/%s' % pulsar.version
ENVIRON_HEADERS = ('content-type', 'content-length')
USER_AGENT = SERVER_SOFTWARE
logger = logging.getLogger('pulsar.proxyserver')


def x_forwarded_for(environ, headers):
    '''Add *x-forwarded-for* header'''
    headers.add_header('x-forwarded-for', environ['REMOTE_ADDR'])


class ProxyServerWsgiHandler(LocalMixin):
    '''WSGI middleware for an asynchronous proxy server.

    To perform processing on headers you can pass a list of
    ``headers_middleware``.
    An headers middleware is a callable which accepts two parameters, the wsgi
    *environ* dictionary and the *headers* container.
    '''
    def __init__(self, headers_middleware=None):
        self.headers_middleware = headers_middleware or []

    @local_property
    def http_client(self):
        '''The :class:`.HttpClient` used by this proxy middleware for
        accessing upstream resources'''
        client = http.HttpClient(decompress=False, store_cookies=False)
        client.headers.clear()
        return client

    def __call__(self, environ, start_response):
        uri = environ['RAW_URI']
        logger.debug('new request for %r' % uri)
        if not uri or uri.startswith('/'):  # No proper uri, raise 404
            raise HttpException(status=404)
        response = TunnelResponse(self, environ, start_response)
        ensure_future(response.request())
        return response.future


############################################################################
#    RESPONSE OBJECTS
class TunnelResponse:
    '''Base WSGI Response Iterator for the Proxy server
    '''
    def __init__(self, wsgi, environ, start_response):
        self.wsgi = wsgi
        self.environ = environ
        self.start_response = start_response
        self.future = create_future()

    async def request(self):
        '''Perform the Http request to the upstream server
        '''
        request_headers = self.request_headers()
        environ = self.environ
        method = environ['REQUEST_METHOD']
        data = None
        if method in ENCODE_BODY_METHODS:
            data = DataIterator(self)
        http = self.wsgi.http_client
        try:
            await http.request(method,
                               environ['RAW_URI'],
                               data=data,
                               headers=request_headers,
                               version=environ['SERVER_PROTOCOL'],
                               pre_request=self.pre_request)
        except Exception as exc:
            self.error(exc)

    def request_headers(self):
        '''Fill request headers from the environ dictionary and
        modify them via the list of :attr:`headers_middleware`.
        The returned headers will be sent to the target uri.
        '''
        headers = Headers(kind='client')
        for k in self.environ:
            if k.startswith('HTTP_'):
                head = k[5:].replace('_', '-')
                headers[head] = self.environ[k]
        for head in ENVIRON_HEADERS:
            k = head.replace('-', '_').upper()
            v = self.environ.get(k)
            if v:
                headers[head] = v
        for middleware in self.wsgi.headers_middleware:
            middleware(self.environ, headers)
        return headers

    def error(self, exc):
        if self.future.done():
            self.future.set_exception(exc)
        else:
            logger.error(str(exc))

    @noerror
    def pre_request(self, response, exc=None):
        '''Start the tunnel.

        This is a callback fired once a connection with upstream server is
        established.
        '''
        if response.request.method == 'CONNECT':
            # proxy - server connection
            upstream = response.connection
            # client - proxy connection
            dostream = self.environ['pulsar.connection']
            # Upgrade downstream connection
            dostream.upgrade(partial(StreamTunnel, upstream))
            # Upgrade upstream connection
            upstream.upgrade(partial(StreamTunnel, dostream))
            self.start_response('200 Connection established', [])
            # send empty byte so that headers are sent
            self.future.set_result([b''])
            response.abort_request()
        else:
            response.bind_event('data_processed', self.data_processed)
            response.bind_event('post_request', self.post_request)

    def data_processed(self, response, data=None, **kw):
        self.environ['pulsar.connection'].write(data)

    def post_request(self, _, exc=None):
        self.future.set_exception(wsgi.AbortWsgi())


class DataIterator:

    def __init__(self, response):
        self.response = response
        self.stream = response.environ.get('wsgi.input')

    def __iter__(self):
        yield self.stream.reader.read()


class StreamTunnel(pulsar.ProtocolConsumer):
    ''':class:`.ProtocolConsumer` handling encrypted messages from
    downstream client and upstream server.

    This consumer is created as an upgrade of the standard Http protocol
    consumer.

    .. attribute:: tunnel

        Connection to the downstream client or upstream server.
    '''
    headers = None
    status_code = None
    http_request = None

    def __init__(self, tunnel, loop=None):
        super().__init__(loop)
        self.tunnel = tunnel

    def connection_made(self, connection):
        self.logger.debug('Tunnel connection %s made', connection)
        connection.bind_event('connection_lost', self._close_tunnel)
        if self.http_request:
            self.start(self.http_request)

    def data_received(self, data):
        try:
            return self.tunnel.write(data)
        except Exception:
            if not self.tunnel.closed:
                raise

    def _close_tunnel(self, arg, exc=None):
        if not self.tunnel.closed:
            self._loop.call_soon(self.tunnel.close)


def server(name='proxy-server', headers_middleware=None,
           server_software=None, **kwargs):
    '''Function to Create a WSGI Proxy Server.'''
    if headers_middleware is None:
        headers_middleware = [x_forwarded_for]
    wsgi_proxy = ProxyServerWsgiHandler(headers_middleware)
    kwargs['server_software'] = server_software or SERVER_SOFTWARE
    return wsgi.WSGIServer(wsgi_proxy, name=name, **kwargs)


if __name__ == '__main__':
    server().start()



import os
import signal
import ctypes
import ctypes.wintypes
import getpass
from multiprocessing import current_process

from .base import *     # noqa

__all__ = ['close_on_exec',
           'daemonize',
           'EXIT_SIGNALS',
           'SIGNALS',
           'kill',
           'get_uid',
           'get_gid',
           'get_maxfd',
           'set_owner_process',
           'current_process']

# See: http://msdn.microsoft.com/en-us/library/ms724935(VS.85).aspx
SetHandleInformation = ctypes.windll.kernel32.SetHandleInformation
SetHandleInformation.argtypes = (ctypes.wintypes.HANDLE, ctypes.wintypes.DWORD,
                                 ctypes.wintypes.DWORD)
SetHandleInformation.restype = ctypes.wintypes.BOOL

MAXFD = 1024
HANDLE_FLAG_INHERIT = 0x00000001
EXIT_SIGNALS = (signal.SIGINT, signal.SIGTERM, signal.SIGABRT, signal.SIGBREAK)
SIGNALS = ()


def kill(pid, sig):
    os.kill(pid, sig)


def set_owner_process(gid, uid):
    return None


def get_parent_id():
    try:
        return os.getppid()
    except Exception:
        return None


def chown(path, uid, gid):
    pass


def close_on_exec(fd):
    if fd:
        success = SetHandleInformation(fd, HANDLE_FLAG_INHERIT, 0)
        if not success:
            raise ctypes.GetLastError()


def _set_non_blocking(fd):
    pass


def get_uid(user=None):
    if not user:
        return getpass.getuser()
    elif user == getpass.getuser():
        return user


def get_gid(group=None):
    return None


def setpgrp():
    pass


def get_maxfd():
    return MAXFD


def daemonize():
    pass



"""
Pulsar :mod:`~greenio` facilitates the integration of synchronous
third-party libraries into pulsar asynchronous framework.
It requires the :greenlet:`greenlet <>` library.

If you want to understand how integration works but you are unfamiliar with
greenlets, check out the :greenlet:`greenlet documentation <>` first.

.. note::

    pulsar-odm_ is a separate library, maintained by the same authors of
    pulsar, which provides asynchronous object data mapping with
    asyncio_ and sqlalchemy_ and uses pulsar greenio extensions.

This application **does not use monkey patching** and therefore it
works quite differently from implicit asynchronous libraries such as
gevent_. This module provides the user with a set
of utilities for **explicitly** transferring execution from one greenlet
to another which executes the blocking call in a greenlet-friendly way.

The caller has the responsibility that the blocking call is greenlet-friendly,
i.e. it transfers the control of execution back to the parent greenlet when
needed.

Usage
=======

Lets assume you are building an application which uses pulsar asynchronous
engine and would like to

* either use an external library written in blocking style,
  i.e. without yielding control to the event loop when IO calls are performed.
* or write your client code without dealing with :class:`~asyncio.Future` or
  coroutines, in other words in an implicit asynchronous style. In this way
  your client code can be used on other frameworks just as well.

In both cases, the :class:`~pulsar.apps.greenio` application is what you need.

.. _green-wsgi:

Green WSGI
---------------

Assume you are using pulsar web server and would like to write your application
in an implicit asynchronous mode, i.e. without dealing with futures nor
coroutines, then you can wrap your WSGI ``app`` with the :class:`.GreenWSGI`
utility::

    from pulsar.apps.wsgi import WSGIServer
    from pulsar.apps.greenio import GreenPool, GreenWSGI

    green_pool = greenio.GreenPool()
    callable = GreenWSGI(app, green_pool)

    WSGIServer(callable=callable).start()

The :class:`.GreenPool` manages a pool of greenlets which execute your
application. In this way, within your ``app`` you can invoke the
:func:`.wait` function when needing to wait for asynchronous results to be
ready.

.. _green-http:

GreenHttp
-----------------

The :class:`.HttpClient` can be used with greenlets::

    >>> from pulsar.apps.greenio import GreenHttp
    >>> http = GreenHttp()

And now you can write synchronous looking code and run it in a separate
greenlet via the :func:`.run_in_greenlet` decorator::

    @greenio.run_in_greenlet
    def example():
        response = http.get('http://bbc.co.uk')
        ...
        return response.text()


and somewhere, in your asynchronous code::

        result = await example()
        result == ...


the :func:`.run_in_greenlet` decorator, execute the function on a child
greenlet without blocking the asynchronous engine. Once the ``example``
function returns, the asynchronous code continue from the ``yield``
statement as usual.


API
======

Wait
----------

.. autofunction:: pulsar.apps.greenio.utils.wait


Run in greenlet
-------------------

.. autofunction:: pulsar.apps.greenio.utils.run_in_greenlet


Green Pool
----------------

.. autoclass:: pulsar.apps.greenio.pool.GreenPool
   :members:
   :member-order: bysource

Green Lock
----------------

.. autoclass:: pulsar.apps.greenio.lock.GreenLock
   :members:
   :member-order: bysource


Green WSGI
----------------

.. autoclass:: pulsar.apps.greenio.wsgi.GreenWSGI
   :members:
   :member-order: bysource


.. _gevent: http://www.gevent.org/
.. _pulsar-odm: https://github.com/quantmind/pulsar-odm
.. _sqlalchemy: http://www.sqlalchemy.org/
.. _asyncio: https://docs.python.org/3/library/asyncio.html
"""
from greenlet import greenlet, getcurrent

from .pool import GreenPool
from .lock import GreenLock
from .http import GreenHttp
from .wsgi import GreenWSGI
from .utils import MustBeInChildGreenlet, wait, run_in_greenlet

__all__ = ['GreenPool',
           'GreenLock',
           'GreenHttp',
           'GreenWSGI',
           'MustBeInChildGreenlet',
           'wait',
           'run_in_greenlet',
           'greenlet',
           'getcurrent']



'''Tests the tools and utilities in pulsar.utils.'''
import unittest
import pickle

from pulsar.utils.structures import (MultiValueDict, merge_prefix, deque,
                                     AttributeDictionary)


class TestMultiValueDict(unittest.TestCase):

    def testConstructor(self):
        m = MultiValueDict()
        self.assertEqual(len(m), 0)
        #
        m = MultiValueDict({'bla': 3})
        self.assertEqual(len(m), 1)
        self.assertEqual(m['bla'], 3)
        #
        m = MultiValueDict({'bla': (3, 78), 'foo': 'ciao'})
        self.assertEqual(len(m), 2)
        self.assertEqual(m['bla'], [3, 78])
        self.assertEqual(m['foo'], 'ciao')
        #
        m = MultiValueDict({'bla': [3, 78], 'foo': (v for v in (1, 2))})
        self.assertEqual(m['bla'], [3, 78])
        self.assertEqual(m['foo'], [1, 2])

    def testset(self):
        m = MultiValueDict()
        m['bla'] = 5
        m['bla'] = 89
        self.assertEqual(m['bla'], [5, 89])
        m['foo'] = 'pippo'
        self.assertEqual(m['foo'], 'pippo')
        return m

    def testextra(self):
        m = MultiValueDict()
        m.setdefault('bla', 'foo')
        self.assertEqual(m['bla'], 'foo')
        m['bla'] = 'ciao'
        self.assertEqual(m['bla'], ['foo', 'ciao'])

    def testget(self):
        m = self.testset()
        self.assertEqual(m.get('sdjcbhjcbh'), None)
        self.assertEqual(m.get('sdjcbhjcbh', 'ciao'), 'ciao')

    def testupdate(self):
        m = self.testset()
        m.update({'bla': 'star', 5: 'bo'})
        self.assertEqual(m['bla'], [5, 89, 'star'])
        self.assertEqual(m[5], 'bo')

    def test_iterators(self):
        m = self.testset()
        d = dict(m.items())
        self.assertEqual(d['bla'], [5, 89])
        self.assertEqual(d['foo'], 'pippo')
        l = list(m.values())
        self.assertEqual(len(l), 2)
        self.assertTrue([5, 89] in l)
        self.assertTrue('pippo' in l)

    def testlists(self):
        m = self.testset()
        items = dict(m.lists())
        self.assertEqual(len(items), 2)
        for k, v in items.items():
            self.assertTrue(isinstance(v, list))
        self.assertEqual(items['bla'], [5, 89])
        self.assertEqual(items['foo'], ['pippo'])

    def testCopy(self):
        m = self.testset()
        m2 = m.copy()
        self.assertEqual(m2['bla'], [5, 89])
        self.assertEqual(m2['foo'], 'pippo')
        self.assertEqual(m2.getlist('foo'), ['pippo'])

    def testPop(self):
        m = self.testset()
        self.assertRaises(KeyError, m.pop, 'jhsdbcjcd')
        self.assertEqual(m.pop('skcbnskcbskcbd', 'ciao'), 'ciao')
        self.assertEqual(m.pop('foo'), 'pippo')
        self.assertRaises(KeyError, m.pop, 'foo')
        self.assertEqual(m.pop('bla'), [5, 89])
        self.assertFalse(m)

    def test_to_dict(self):
        m = MultiValueDict((('id', 1), ('id', 2)))
        self.assertEqual(len(m), 1)
        self.assertEqual(m['id'], [1, 2])
        d = dict(m)
        self.assertEqual(d, {'id': [1, 2]})


class TestAttributeDictionary(unittest.TestCase):

    def testInit(self):
        self.assertRaises(TypeError, AttributeDictionary, {}, {})
        a = AttributeDictionary({'bla': 1}, foo='pippo')
        self.assertEqual(dict(a), {'bla': 1, 'foo': 'pippo'})
        self.assertEqual(len(a), 2)

    def testAssign(self):
        a = AttributeDictionary()
        a['ciao'] = 5
        self.assertEqual(a.ciao, 5)
        self.assertEqual(a['ciao'], 5)
        self.assertEqual(list(a.values()), [5])
        self.assertEqual(list(a.items()), [('ciao', 5)])

    def testCopy(self):
        a = AttributeDictionary(foo=5, bla='ciao')
        self.assertEqual(len(a), 2)
        b = a.copy()
        self.assertEqual(a, b)
        self.assertNotEqual(id(a), id(b))

    def __test_pickle(self):
        # TODO: this fails at times
        a = AttributeDictionary()
        a['ciao'] = 5
        b = pickle.dumps(a)
        c = pickle.loads(b)
        self.assertEqual(a, c)


class TestFunctions(unittest.TestCase):

    def test_merge_prefix(self):
        d = deque([b'abc', b'de', b'fghi', b'j'])
        merge_prefix(d, 5)
        self.assertEqual(d, deque([b'abcde', b'fghi', b'j']))
        d = deque([b'abc', b'de', b'fghi', b'j'])
        merge_prefix(d, 4)
        self.assertEqual(d, deque([b'abcd', b'e', b'fghi', b'j']))
        merge_prefix(d, 7)
        self.assertEqual(d, deque([b'abcdefg', b'hi', b'j']))
        merge_prefix(d, 3)
        self.assertEqual(d, deque([b'abc', b'defg', b'hi', b'j']))
        merge_prefix(d, 100)
        self.assertEqual(d, deque([b'abcdefghij']))



'''Tests the test suite loader.'''
import os
import unittest

from pulsar import get_application


class TestTestLoader(unittest.TestCase):

    async def test_testsuite(self):
        app = await get_application('test')
        self.assertTrue(app.cfg.script)
        # self.assertEqual(app.script, sys.argv[0])
        self.assertEqual(os.path.dirname(app.cfg.script), app.root_dir)
        self.assertEqual(app.cfg.test_modules, ['tests', 'examples'])

    async def test_load_pulsar_tests(self):
        app = await get_application('test')
        modules = dict(app.loader.test_files())
        self.assertTrue(modules)
        self.assertFalse('httpbin' in modules)
        self.assertTrue('echo' in modules)
        self.assertFalse('async' in modules)

    async def test_sorted_tags(self):
        app = await get_application('test')
        modules = app.loader.test_files()
        self.assertTrue(modules)
        tags = [m[0] for m in modules]
        self.assertEqual(tags, sorted(tags))

    async def test_load_tags1(self):
        app = await get_application('test')
        modules = dict(app.loader.test_files(('suite',)))
        self.assertEqual(len(modules), 4)

    async def test_load_exclude(self):
        app = await get_application('test')
        modules = dict(app.loader.test_files(
            exclude=('stores', 'apps.pubsub')))
        self.assertFalse('stores.lock' in modules)
        self.assertFalse('stores.parser' in modules)
        self.assertFalse('apps.pubsub' in modules)
        self.assertTrue('apps.greenio' in modules)

    async def test_load_test_function(self):
        app = await get_application('test')
        modules = dict(app.loader.test_files(
            ['suite.loader.test_load_test_function']))
        self.assertEqual(len(modules), 1)
        self.assertTrue('suite.loader' in modules)
        self.assertEqual(modules['suite.loader'][1], 'test_load_test_function')

    async def test_load_bad_test_function(self):
        app = await get_application('test')
        modules = dict(app.loader.test_files(
            ['suite.loader.load_test_function']))
        self.assertEqual(len(modules), 0)



import sys
import threading
import logging
from collections import deque

from pulsar import create_future, ensure_future, AsyncObject, get_event_loop

from .utils import wait, GreenletWorker, isawaitable, getcurrent


_DEFAULT_WORKERS = 100
_MAX_WORKERS = 1000


class _DONE:
    pass


class GreenPool(AsyncObject):
    """A pool of running greenlets.

    This pool maintains a group of greenlets to perform asynchronous
    tasks via the :meth:`submit` method.
    """
    worker_name = 'exec'

    def __init__(self, max_workers=None, loop=None):
        self._loop = loop or get_event_loop()
        self._max_workers = min(max_workers or _DEFAULT_WORKERS, _MAX_WORKERS)
        self._greenlets = set()
        self._available = set()
        self._queue = deque()
        self._shutdown = False
        self._waiter = None
        self._logger = logging.getLogger('pulsar.greenpool')
        self._shutdown_lock = threading.Lock()
        self.wait = wait

    @property
    def max_workers(self):
        return self._max_workers

    @max_workers.setter
    def max_workers(self, value):
        value = int(value)
        assert value > 0
        self._max_workers = value

    @property
    def in_green_worker(self):
        """True if the current greenlet is a green pool worker
        """
        return isinstance(getcurrent(), GreenletWorker)

    @property
    def closed(self):
        """True if this pool is closed and no task can queued
        """
        return self._shutdown

    def submit(self, func, *args, **kwargs):
        """Equivalent to ``func(*args, **kwargs)``.

        This method create a new task for function ``func`` and adds it to
        the queue.
        Return a :class:`~asyncio.Future` called back once the task
        has finished.
        """
        with self._shutdown_lock:
            if self._shutdown:
                raise RuntimeError(
                    'cannot schedule new futures after shutdown')
            if self.in_green_worker:
                return wait(func(*args, **kwargs))
            else:
                future = create_future(self._loop)
                self._put((future, func, args, kwargs))
                return future

    def shutdown(self, wait=True):
        with self._shutdown_lock:
            self._shutdown = True
            self._put(None)
            if wait:
                self._waiter = create_future(self._loop)
                return self._waiter

    def getcurrent(self):
        return getcurrent()

    # INTERNALS
    def _adjust_greenlet_count(self):
        if (not self._shutdown and not self._available and
                len(self._greenlets) < self._max_workers):
            green = GreenletWorker(self._green_run)
            self._greenlets.add(green)
            self.logger.debug('Num greenlets: %d', len(self._greenlets))
            green.switch()
        return self._available

    def _put(self, task):
        # Run in the main greenlet of the evnet-loop thread
        self._queue.appendleft(task)
        self._check_queue()

    def _check_queue(self):
        # Run in the main greenlet of the event-loop thread
        if not self._adjust_greenlet_count():
            self.logger.debug('No greenlet available')
            return self._loop.call_soon(self._check_queue)
        try:
            task = self._queue.pop()
        except IndexError:
            return
        ensure_future(self._green_task(self._available.pop(), task),
                      loop=self._loop)

    async def _green_task(self, green, task):
        # Coroutine executing the in main greenlet
        # This coroutine is executed for every task put into the queue

        while task is not _DONE:
            # switch to the greenlet to start the task
            task = green.switch(task)

            # if an asynchronous result is returned, await
            while isawaitable(task):
                try:
                    task = await task
                except Exception as exc:
                    # This call can return an asynchronous component
                    exc_info = sys.exc_info()
                    if not exc_info[0]:
                        exc_info = (exc, None, None)
                    task = green.throw(*exc_info)

    def _green_run(self):
        # The run method of a worker greenlet
        task = True
        while task:
            green = getcurrent()
            parent = green.parent
            assert parent
            # add greenlet in the available greenlets
            self._available.add(green)
            task = parent.switch(_DONE)  # switch back to the main execution
            if task:
                future, func, args, kwargs = task
                try:
                    try:
                        result = wait(func(*args, **kwargs), True)
                    except StopIteration as exc:  # See PEP 479
                        raise RuntimeError('Unhandled StopIteration') from exc
                except Exception as exc:
                    future.set_exception(exc)
                else:
                    future.set_result(result)
            else:  # Greenlet cleanup
                self._greenlets.remove(green)
                if self._greenlets:
                    self._put(None)
                elif self._waiter:
                    self._waiter.set_result(None)
                    self._waiter = None
                parent.switch(_DONE)



'''Tests the "flaskgreen" example.'''
import unittest

from pulsar import send, SERVER_SOFTWARE
from pulsar.apps.http import HttpClient
from pulsar.apps.test import dont_run_with_thread

try:
    from examples.flaskgreen.manage import FlaskGreen, log_connection
except ImportError:
    FlaskGreen = None


@unittest.skipUnless(FlaskGreen, "Requires flask and greenlet module")
class TestFlaskGreenThread(unittest.TestCase):
    app_cfg = None
    concurrency = 'thread'

    @classmethod
    def name(cls):
        return 'flaskgreen_' + cls.concurrency

    @classmethod
    async def setUpClass(cls):
        s = FlaskGreen(name=cls.name(),
                       concurrency=cls.concurrency,
                       bind='127.0.0.1:0',
                       echo_bind='127.0.0.1:0')
        cls.app_cfg = await send('arbiter', 'run', s)
        cls.uri = 'http://{0}:{1}'.format(*cls.app_cfg[0].addresses[0])
        cls.client = HttpClient()

    @classmethod
    async def tearDownClass(cls):
        if cls.app_cfg is not None:
            await send('arbiter', 'kill_actor', cls.app_cfg[0].name)
            await send('arbiter', 'kill_actor', cls.app_cfg[1].name)

    def test_apps(self):
        self.assertEqual(self.app_cfg[0].name, self.name())
        self.assertEqual(self.app_cfg[1].name, 'echo_%s' % self.name())
        self.assertEqual(self.app_cfg[1].connection_made, log_connection)

    async def testResponse200(self):
        c = self.client
        response = await c.get(self.uri)
        self.assertEqual(response.status_code, 200)
        content = response.content
        self.assertEqual(content, b'Try any other url for an echo')
        headers = response.headers
        self.assertTrue(headers)
        self.assertEqual(headers['server'], SERVER_SOFTWARE)

    async def testResponseEcho(self):
        c = self.client
        response = await c.get('%s/ciao' % self.uri)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b'Flask Url : ciao')


@dont_run_with_thread
class TestFlaskGreenProcess(TestFlaskGreenThread):
    concurrency = 'process'



"""This is the most important :ref:`pulsar application <application-api>`.
The server is a specialized :ref:`socket server <apps-socket>`
for web applications conforming with the python web server
gateway interface (`WSGI 1.0.1`_).
The server can be used in conjunction with several web frameworks
as well as :ref:`pulsar wsgi application handlers <wsgi-handlers>`,
:ref:`pulsar router <wsgi-middleware>`,
the :ref:`pulsar RPC middleware <apps-rpc>` and
the :ref:`websocket middleware <apps-ws>`.

.. note::

    Pulsar wsgi server is production ready designed to easily
    handle fast, scalable http applications. As all pulsar applications,
    it uses an event-driven, non-blocking I/O model that makes it
    lightweight and efficient. In addition, its multiprocessing
    capabilities allow to handle the `c10k problem`_ with ease.


An example of a web server written with the :mod:`~pulsar.apps.wsgi`
module which responds with ``Hello World!`` for every request::

    from pulsar.apps import wsgi

    def hello(environ, start_response):
        data = b"Hello World!"
        response_headers = [('Content-type','text/plain'),
                            ('Content-Length', str(len(data)))]
        start_response("200 OK", response_headers)
        return [data]

    if __name__ == '__main__':
        wsgi.WSGIServer(hello).start()


For more information regarding WSGI check the pep3333_ specification.
To run the application::

    python script.py

For available run options::

    python script.py --help



WSGI Server
===================

.. autoclass:: WSGIServer
   :members:
   :member-order: bysource


.. _`WSGI 1.0.1`: http://www.python.org/dev/peps/pep-3333/
.. _`c10k problem`: http://en.wikipedia.org/wiki/C10k_problem
"""
from functools import partial

import pulsar
from pulsar.apps.socket import SocketServer, Connection

from .html import HtmlVisitor
from .content import (String, Html, Json, HtmlDocument, Links, Scripts, Media,
                      html_factory)
from .middleware import (clean_path_middleware, authorization_middleware,
                         wait_for_body_middleware, middleware_in_executor)
from .response import AccessControl, GZipMiddleware
from .wrappers import EnvironMixin, WsgiResponse, WsgiRequest, cached_property
from .server import HttpServerResponse, test_wsgi_environ, AbortWsgi
from .route import route, Route
from .handlers import WsgiHandler, LazyWsgi
from .routers import (Router, MediaRouter, MediaMixin, RouterParam,
                      file_response)
from .auth import HttpAuthenticate, parse_authorization_header
from .formdata import parse_form_data
from .utils import (handle_wsgi_error, render_error_debug, wsgi_request,
                    set_wsgi_request_class, dump_environ, HOP_HEADERS)

__all__ = [
    # Server
    'WSGIServer',
    'HttpServerResponse',
    'test_wsgi_environ',
    'AbortWsgi',
    #
    # Content strings
    'String',
    'Html',
    'Json',
    'HtmlDocument',
    'Links',
    'Scripts',
    'Media',
    'html_factory',
    'HtmlVisitor',
    #
    # Request middleware
    'clean_path_middleware',
    'authorization_middleware',
    'wait_for_body_middleware',
    'middleware_in_executor',
    #
    # Response middleware
    'AccessControl',
    'GZipMiddleware',
    #
    # WSGI Wrappers
    'EnvironMixin',
    'WsgiResponse',
    'WsgiRequest',
    'cached_property',
    #
    # WSGI Handlers
    'WsgiHandler',
    'LazyWsgi',
    #
    # Routes and Routers
    'route',
    'Route',
    'Router',
    'MediaRouter',
    'MediaMixin',
    'RouterParam',
    'file_response',
    #
    # Utilities
    'parse_form_data',
    'HttpAuthenticate',
    'parse_authorization_header',
    'handle_wsgi_error',
    'render_error_debug',
    'wsgi_request',
    'set_wsgi_request_class',
    'dump_environ',
    'HOP_HEADERS'
]


class WSGIServer(SocketServer):
    '''A WSGI :class:`.SocketServer`.
    '''
    name = 'wsgi'
    cfg = pulsar.Config(apps=['socket'],
                        server_software=pulsar.SERVER_SOFTWARE)

    def protocol_factory(self):
        cfg = self.cfg
        consumer_factory = partial(HttpServerResponse, cfg.callable, cfg,
                                   cfg.server_software)
        return partial(Connection, consumer_factory)



#!/usr/bin/env python
import sys
import os
import json
import subprocess

from setuptools import setup, find_packages
from extensions import ext

import pulsar_test


class PulsarTest(pulsar_test.Test):
    start_coverage = True


def extend(params, package=None):
    if package:
        path = os.path.dirname(__file__)
        data = sh('%s %s package_info %s %s'
                  % (sys.executable, __file__, package, path))
        params.update(json.loads(data))

    return params


def read(name):
    filename = os.path.join(os.path.dirname(__file__), name)
    with open(filename) as fp:
        return fp.read()


def requirements(name):
    install_requires = []
    dependency_links = []

    for line in read(name).split('\n'):
        if line.startswith('-e '):
            link = line[3:].strip()
            if link == '.':
                continue
            dependency_links.append(link)
            line = link.split('=')[1]
        line = line.strip()
        if line:
            install_requires.append(line)

    return install_requires, dependency_links


def sh(command):
    proc = subprocess.Popen(command,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            shell=True,
                            universal_newlines=True)
    stdout, stderr = proc.communicate()
    if proc.returncode:
        if stdout:
            sys.stdout.write(stdout)
        if stderr:
            sys.stderr.write(stderr)
        exit(proc.returncode)
    return stdout


meta = dict(
    name='pulsar',
    author="Luca Sbardella",
    author_email="luca@quantmind.com",
    maintainer_email="luca@quantmind.com",
    url="https://github.com/quantmind/pulsar",
    license="BSD",
    long_description=read('README.rst'),
    include_package_data=True,
    setup_requires=['wheel'],
    packages=find_packages(include=['pulsar', 'pulsar.*', 'pulsar_test']),
    entry_points={
        "distutils.commands": [
            "pulsar_test = pulsar_test:Test"
        ]
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet',
        'Topic :: Utilities',
        'Topic :: System :: Distributed Computing',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Server',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content']
)


def run_setup(with_cext):
    params = ext.params() if with_cext else {}
    params.update(meta)
    cmdclass = params.get('cmdclass', {})
    cmdclass['test'] = PulsarTest
    cmdclass['bench'] = pulsar_test.Bench
    params['cmdclass'] = cmdclass
    setup(**extend(params, 'pulsar'))


def package_info():
    package = sys.argv[2]
    if len(sys.argv) > 3:
        sys.path.append(sys.argv[3])
    os.environ['package_info'] = package
    pkg = __import__(package)
    print(json.dumps(dict(version=pkg.__version__,
                          description=pkg.__doc__)))


if __name__ == '__main__':
    command = sys.argv[1] if len(sys.argv) > 1 else None
    if command == 'package_info':
        package_info()
    elif command == 'agile':
        from agile.app import AgileManager
        AgileManager(description='Release manager for pulsar',
                     argv=sys.argv[2:]).start()
    else:
        try:
            run_setup(True)
        except ext.BuildFailed as exc:
            print('WARNING: C extensions could not be compiled: %s' % exc.msg)
            run_setup(False)



from random import randint
import unittest

from pulsar.utils.websocket import frame_parser


def i2b(args):
    return bytes(bytearray(args))


class TestCParser(unittest.TestCase):
    __benchmark__ = True
    __number__ = 10000
    _sizes = {'tiny': 10,
              'small': 200,
              'normal': 2000,
              'large': 10000,
              'huge': 100000}

    @classmethod
    def setUpClass(cls):
        size = cls.cfg.size
        nsize = cls._sizes[size]
        cls.data = i2b((randint(0, 255) for v in range(nsize)))

    def setUp(self):
        self.server = self.parser()
        self.client = self.parser(kind=1)

    def parser(self, kind=0):
        return frame_parser(kind=kind)

    def test_masked_encode(self):
        self.client.encode(self.data, opcode=2)


class TestPyParser(TestCParser):

    def parser(self, kind=0):
        return frame_parser(pyparser=True, kind=kind)



from random import randint
import unittest

import pulsar
from pulsar.apps.ds import (redis_parser, ResponseError, NoScriptError,
                            InvalidResponse)


def lua_nested_table(nesting, s=100):
    s = b''.join((b'1234567890' for n in range(s)))
    pres = [100, s]
    result = pres
    for i in range(nesting):
        res = [-8, s, None]
        # dres = {1: 'valid', 'bla': 6}
        dres = 1
        pres.extend((res, res, dres))
        pres = res
    return result


class TestParser(unittest.TestCase):

    def parser(self):
        return redis_parser()()

    #    DECODER
    def test_null(self):
        test = b'$-1\r\n'
        p = self.parser()
        p.feed(test)
        self.assertEqual(p.get(), None)

    def test_empty_string(self):
        test = b'$0\r\n\r\n'
        p = self.parser()
        p.feed(test)
        self.assertEqual(p.get(), b'')
        self.assertEqual(p.buffer(), b'')

    def test_empty_vector(self):
        test = b'*0\r\n'
        p = self.parser()
        p.feed(test)
        self.assertEqual(p.get(), [])
        self.assertEqual(p.buffer(), b'')

    def test_parseError(self):
        test = b'pxxxx\r\n'
        p = self.parser()
        p.feed(test)
        self.assertRaises(InvalidResponse, p.get)

    def test_responseError(self):
        test = b'-ERR random error\r\n'
        p = self.parser()
        p.feed(test)
        value = p.get()
        self.assertIsInstance(value, ResponseError)
        self.assertEqual(str(value), 'random error')

    def test_noscriptError(self):
        test = b'-NOSCRIPT random error\r\n'
        p = self.parser()
        p.feed(test)
        value = p.get()
        self.assertIsInstance(value, NoScriptError)
        self.assertEqual(str(value), 'random error')

    def test_binary(self):
        test = (b'$31\r\n\x80\x02]q\x00(X\x04\x00\x00\x00ciaoq\x01X\x05\x00'
                b'\x00\x00pippoq\x02e.\r\n')
        p = self.parser()
        p.feed(test)
        self.assertEqual(p.buffer(), test)
        value = p.get()
        self.assertTrue(value)
        self.assertEqual(p.buffer(), b'')

    def test_multi(self):
        test = (b'+OK\r\n+QUEUED\r\n+QUEUED\r\n+QUEUED\r\n*3\r\n$-1\r\n:1'
                b'\r\n:39\r\n')
        p = self.parser()
        p.feed(test)
        self.assertEqual(p.get(), b'OK')
        self.assertEqual(p.get(), b'QUEUED')
        self.assertEqual(p.get(), b'QUEUED')
        self.assertEqual(p.get(), b'QUEUED')
        self.assertEqual(p.get(), [None, 1, 39])

    def test_nested(self):
        p = self.parser()
        result = lua_nested_table(2)
        chunk = p.multi_bulk(result)
        p.feed(chunk)
        res2 = p.get()
        self.assertEqual(len(res2), len(result))
        self.assertEqual(res2[0], b'100')
        self.assertEqual(res2[1], result[1])

    def test_nested_bug_fixer(self):
        p = self.parser()
        result = [b'100',
                  b'1234567890',
                  [b'-8', b'1234567890', None],
                  [b'-8', b'1234567890', None],
                  b'1']
        data = p.multi_bulk(result)
        chunks = [b'*5\r\n$3\r\n100',
                  b'\r\n$10\r\n12345',
                  b'67890\r\n*3',
                  b'\r\n$2\r',
                  b'\n-8\r\n$10\r',
                  b'\n1234567890\r\n$-1\r\n',
                  b'*3\r\n$2\r\n-8\r\n$10\r\n1234567890\r\n$-1\r\n$1\r\n1\r\n']
        self.assertEqual(data, b''.join(chunks))
        for chunk in chunks[:-1]:
            p.feed(chunk)
            self.assertEqual(p.get(), False)
        p.feed(chunks[-1])
        self.assertEqual(p.get(), result)

    def test_nested_a_drop_at_a_time(self):
        p = self.parser()
        result = lua_nested_table(3)
        data = p.multi_bulk(result)
        while data:
            self.assertEqual(p.get(), False)
            i = randint(1, 30)
            chunk, data = data[:i], data[i:]
            p.feed(chunk)
        res2 = p.get()
        self.assertEqual(len(res2), len(result))
        self.assertEqual(res2[0], b'100')
        self.assertEqual(res2[1], result[1])

    # CLIENT ENCODERS
    def test_encode_commands(self):
        p = self.parser()
        chunk = p.pack_command(['whatever', None])
        self.assertEqual(chunk, b'*2\r\n$8\r\nwhatever\r\n$4\r\nNone\r\n')

    # SERVER ENCODERS

    def test_encode_empty_string(self):
        p = self.parser()
        chunk = p.bulk(None)
        self.assertEqual(chunk, b'$-1\r\n')

    def test_encode_multi_bulk(self):
        p = self.parser()
        self.assertEqual(p.multi_bulk([]), b'*0\r\n')
        self.assertEqual(p.multi_bulk(()), b'*0\r\n')


@unittest.skipUnless(pulsar.HAS_C_EXTENSIONS, 'Requires C extensions')
class TestPythonParser(TestParser):

    def parser(self):
        return redis_parser(True)()



from pulsar import Http404
from pulsar.apps.wsgi import handle_wsgi_error, WsgiResponse

from .utils import wait


class GreenWSGI:
    '''Wraps a WSGI application to be executed on a :class:`.GreenPool`
    '''
    def __init__(self, middleware, pool, response_middleware=None):
        if not isinstance(middleware, (list, tuple)):
            middleware = [middleware]
        self.middleware = list(middleware)
        self.response_middleware = response_middleware or []
        self.pool = pool

    def __call__(self, environ, start_response):
        if self.pool.in_green_worker:
            return self._call(environ, start_response)
        else:
            return self.pool.submit(self._call, environ, start_response)

    def _call(self, environ, start_response):
        wsgi_input = environ['wsgi.input']
        if wsgi_input and not isinstance(wsgi_input, GreenStream):
            environ['wsgi.input'] = GreenStream(wsgi_input)
        response = None
        try:
            for middleware in self.middleware:
                response = wait(middleware(environ, start_response))
                if response is not None:
                    break
            if response is None:
                raise Http404

        except Exception as exc:
            response = wait(handle_wsgi_error(environ, exc))

        if isinstance(response, WsgiResponse) and not response.started:
            for middleware in self.response_middleware:
                response = wait(middleware(environ, response)) or response
            response.start(start_response)
        return response


class GreenStream:
    __slots__ = ('stream',)

    def __init__(self, stream):
        self.stream = stream

    def __getattr__(self, name):
        value = getattr(self.stream, name)
        if getattr(value, '__self__', None) is self.stream:
            return _green(value)
        return value


class _green:
    __slots__ = ('value',)

    def __init__(self, value):
        self.value = value

    def __getattr__(self, name):
        return getattr(self.value, name)

    def __call__(self, *args, **kwargs):
        return wait(self.value(*args, **kwargs))



import unittest

from pulsar.utils.path import Path


class TestPath(unittest.TestCase):

    def testThis(self):
        p = Path(__file__)
        self.assertTrue(p.isfile())
        self.assertFalse(p.isdir())
        c = Path.cwd()
        self.assertNotEqual(p, c)
        self.assertTrue(c.isdir())

    def testDir(self):
        c = Path.cwd()
        self.assertEqual(c, c.dir())
        c = Path('/sdjc/scdskjcdnsd/dhjdhjdjksdjksdksd')
        self.assertFalse(c.exists())
        self.assertRaises(ValueError, c.dir)

    def testAdd2Python(self):
        p = Path('/sdjc/scdskjcdnsd/dhjdhjdjksdjksdksd')
        module = p.add2python('pulsar')
        self.assertEqual(module.__name__, 'pulsar')
        self.assertRaises(ValueError, p.add2python, 'kaputttt')

    def testAdd2Python_failure(self):
        p = Path()
        self.assertRaises(ImportError, p.add2python, 'kaputttt')
        self.assertFalse(p.add2python('vdfvdavfdv', must_exist=False))



import unittest

from pulsar.apps.wsgi import WsgiRequest


class AcceptTests(unittest.TestCase):

    def test_empty_mime(self):
        environ = {}
        a = WsgiRequest(environ)
        content_types = a.content_types
        self.assertFalse(content_types)

    def test_content_types(self):
        environ = {'HTTP_ACCEPT': ('text/html, application/xhtml+xml, '
                                   'application/xml;q=0.9, */*;q=0.8')}
        a = WsgiRequest(environ)
        content_types = a.content_types
        self.assertTrue(content_types)
        self.assertEqual(len(content_types), 4)
        self.assertEqual(id(content_types), id(a.content_types))
        self.assertTrue('text/html' in content_types)
        self.assertTrue('text/plain' in content_types)
        self.assertEqual(content_types.quality('text/html'), 1)
        self.assertEqual(content_types.quality('text/plain'), 0.8)
        self.assertEqual(content_types.quality('application/json'), 0.8)
        self.assertEqual(content_types.best, 'text/html')

    def test_best(self):
        environ = {'HTTP_ACCEPT': ('text/html, application/xhtml+xml, '
                                   'application/xml;q=0.9, */*;q=0.8')}
        a = WsgiRequest(environ)
        content_types = a.content_types
        self.assertTrue(content_types)
        self.assertEqual(content_types.best_match(('text/html',
                                                   'application/json')),
                         'text/html')
        self.assertEqual(content_types.best_match(('application/json',
                                                   'text/html')),
                         'text/html')



import unittest

from pulsar import create_future, get_event_loop


DELAY = 0


def async_func(loop, value):
    p = create_future(loop)
    loop.call_later(DELAY, p.set_result, value)
    return p


async def sub_sub(loop, num):
    a = await async_func(loop, num)
    b = await async_func(loop, num)
    return a+b


async def sub(loop, num):
    return (
        await async_func(loop, num) +
        await async_func(loop, num) +
        await sub_sub(loop, num)
    )

async def main(num, loop=None):
    loop = loop or get_event_loop()
    a = await async_func(loop, num)
    b = await sub(loop, num)
    c = await sub(loop, num)
    return a+b+c


class TestCoroutine(unittest.TestCase):
    __benchmark__ = True
    __number__ = 1000

    async def test_coroutine(self):
        result = await main(1)
        self.assertEqual(result, 9)



try:
    from oauthlib import oauth1, oauth2
    from oauthlib.common import generate_token
except ImportError:
    oauth1 = None
    oauth2 = None

from pulsar import ImproperlyConfigured
from pulsar.utils.structures import mapping_iterator

from . import auth


class OAuth1(auth.Auth):
    '''Add OAuth1 authentication to pulsar :class:`.HttpClient`
    '''

    def __init__(self, client_id=None, client=None, **kw):
        if oauth1 is None:
            raise ImproperlyConfigured('%s requires oauthlib' %
                                       self.__class__.__name__)
        self._client = client or oauth1.Client(client_id, **kw)

    def __call__(self, response, exc=None):
        r = response.request
        url, headers, _ = self._client.sign(
            r.url, r.method, r.body, r.headers)
        for key, value in mapping_iterator(headers):
            r.add_header(key, value)
        r.url = url


class OAuth2(auth.Auth):
    '''Add OAuth2 authentication to pulsar :class:`.HttpClient`'''

    def __init__(self, client_id=None, client=None, **kw):
        if oauth2 is None:
            raise ImproperlyConfigured('%s requires oauthlib' %
                                       self.__class__.__name__)
        self.client = client or oauth2.WebApplicationClient(client_id, **kw)

    def __call__(self, response, exc=None):
        r = response.request
        url, headers, _ = self.client.add_token(
            r.url, http_method=r.method, body=r.body, headers=r.headers)
        for key, value in mapping_iterator(headers):
            r.add_header(key, value)
        r.url = url

    def prepare_request_uri(self, url, state=None, **kwargs):
        """Prepare the request uri and return a tuple with url and state

        :param url: base url for OAuth2 requests
        :param state: optional state, if not supplied one is generated
        :param kwargs: additional parameters
        :return: a tuple with url and state
        """
        state = state or generate_token()
        url = self.client.prepare_request_uri(url, state=state, **kwargs)
        return url, state



import pulsar
from pulsar.apps.test import ActorTestMixin
from pulsar import send, async_while

from tests.async import (add, get_test, spawn_actor_from_actor, close_mailbox,
                         wait_for_stop, check_environ)


class ActorTest(ActorTestMixin):

    async def test_spawn_and_interact(self):
        name = 'pluto-%s' % self.concurrency
        proxy = await self.spawn_actor(name=name)
        self.assertEqual(proxy.name, name)
        self.assertEqual(await send(proxy, 'ping'), 'pong')
        self.assertEqual(await send(proxy, 'echo', 'Hello!'), 'Hello!')
        n, result = await send(proxy, 'run', add, 1, 3)
        self.assertEqual(n, name)
        self.assertEqual(result, 4)

    async def test_info(self):
        name = 'pippo-%s' % self.concurrency
        proxy = await self.spawn_actor(name=name)
        self.assertEqual(proxy.name, name)
        info = await send(proxy, 'info')
        self.assertTrue('actor' in info)
        ainfo = info['actor']
        self.assertEqual(ainfo['is_process'],
                         self.concurrency in ('process', 'multi'))

    async def test_simple_spawn(self):
        '''Test start and stop for a standard actor on the arbiter domain.'''
        proxy = await self.spawn_actor(
            name='simple-actor-on-%s' % self.concurrency)
        arbiter = pulsar.get_actor()
        proxy_monitor = arbiter.get_actor(proxy.aid)
        self.assertEqual(proxy_monitor, proxy)
        self.assertEqual(await send(proxy, 'ping'), 'pong')
        self.assertEqual(await send(proxy.proxy, 'echo', 'Hello!'), 'Hello!')
        # We call the ActorTestMixin.stop_actors method here, since the
        # ActorTestMixin.tearDown method is invoked on the test-worker domain
        # (here we are in the arbiter domain)
        await self.stop_actors(proxy)
        is_alive = await async_while(3, proxy_monitor.is_alive)
        self.assertFalse(is_alive)

    async def test_spawn_from_actor(self):
        proxy = await self.spawn_actor(
            name='spawning-actor-%s' % self.concurrency)
        arbiter = pulsar.get_actor()
        self.assertTrue(repr(proxy).startswith('spawning-actor-'))
        self.assertEqual(proxy, proxy.proxy)
        proxy_monitor = arbiter.get_actor(proxy.aid)
        self.assertFalse(proxy != proxy_monitor)
        #
        # do the spawning
        name = 'spawned-actor-%s-from-actor' % self.concurrency
        aid = await send(proxy, 'run', spawn_actor_from_actor, name)
        self.assertTrue(aid)
        proxy_monitor2 = arbiter.get_actor(aid)
        self.assertEqual(proxy_monitor2.name, name)
        self.assertNotEquals(proxy_monitor, proxy_monitor2)
        #
        # stop them
        await self.stop_actors(proxy, proxy_monitor2)
        is_alive = await async_while(3, proxy_monitor.is_alive)
        self.assertFalse(is_alive)
        is_alive = await async_while(3, proxy_monitor2.is_alive)
        self.assertFalse(is_alive)

    async def test_config_command(self):
        proxy = await self.spawn_actor(
            name='actor-test-config-%s' % self.concurrency)
        arbiter = pulsar.get_actor()
        proxy_monitor = arbiter.get_actor(proxy.aid)
        result = await send(proxy, 'config', 'khjkh', 'name')
        self.assertEqual(result, None)
        result = await send(proxy, 'config', 'get', 'concurrency')
        self.assertEqual(result, self.concurrency)
        result = await send(proxy, 'config', 'get', 'concurrency', 'foo')
        self.assertEqual(result, None)
        #
        result = await send(proxy, 'config', 'set', 'max_requests', 1000, 1000)
        self.assertEqual(result, None)
        result = await send(proxy, 'config', 'set', 'max_requests', 1000)
        self.assertEqual(result, True)
        result = await send(proxy, 'config', 'get', 'max_requests')
        self.assertEqual(result, 1000)
        #
        await self.stop_actors(proxy)
        is_alive = await async_while(3, proxy_monitor.is_alive)
        self.assertFalse(is_alive)

    async def test_get_application(self):
        proxy = await self.spawn_actor(
            name='actor-test-get-application-%s' % self.concurrency)
        cfg = await send(proxy, 'run', get_test)
        self.assertTrue(cfg)
        self.assertEqual(cfg.name, 'test')

    async def test_connection_lost(self):
        proxy = await self.spawn_actor(
            name='actor-test-connection-lost-%s' % self.concurrency)
        await send(proxy, 'run', close_mailbox)
        await wait_for_stop(self, proxy.aid, True)

    async def test_environment(self):
        import os
        os.environ['PULSAR_TEST_ENVIRON'] = 'yes'
        name = 'environment-%s' % self.concurrency
        proxy = await self.spawn_actor(name=name)
        value = await send(proxy, 'run', check_environ, 'PULSAR_TEST_ENVIRON')
        self.assertEqual(value, 'yes')



"""This script demonstrates the use synchronous web frameworks such as flask_
in the context of asynchronous responses. It uses pulsar the :mod:`.greenio`
module to run flask on a pool of greenlet.

This configuration allows the use of the greenlet-friendly :class:`EchoGreen`
client within a flask response function.
The Echo server starts at the same time as the Flask WSGI server.

To run the script::

    python manage.py

and point to http://localhost:8080 or any other sub urls.

Main Classes
==============

.. autoclass:: FlaskSite

.. autoclass:: EchoGreen

.. autoclass:: FlaskSite

Components
=============

In this snippet the following components are used:

* Pulsar :class:`.MultiApp`
* Pulsar :class:`.WSGIServer`
* Pulsar :class:`.SocketServer`
* Pulsar :class:`.GreenWSGI`

For more information about writing server and clients check the
:ref:`writing clients tutorial <tutorials-writing-clients>`.

.. _flask: http://flask.pocoo.org/
"""
from functools import partial

from flask import Flask, make_response

from pulsar.apps import wsgi

import pulsar
from pulsar import Pool, Connection, AbstractClient, ProtocolError
from pulsar.apps.socket import SocketServer

from pulsar import MultiApp, Config
from pulsar.apps.wsgi import WSGIServer

from pulsar.apps.greenio import GreenPool
from pulsar.apps.greenio.wsgi import GreenWSGI


class EchoProtocol(pulsar.ProtocolConsumer):
    separator = b'\r\n\r\n'
    '''A separator for messages.'''
    buffer = b''
    '''The buffer for long messages'''

    def data_received(self, data):
        if self.buffer:
            data = self.buffer + data

        idx = data.find(self.separator)
        if idx >= 0:    # we have a full message
            idx += len(self.separator)
            data, rest = data[:idx], data[idx:]
            self.buffer = self.response(data, rest)
            self.finished()
            return rest
        else:
            self.buffer = data

    def start_request(self):
        self.transport.write(self._request + self.separator)

    def response(self, data, rest):
        if rest:
            raise ProtocolError
        return data[:-len(self.separator)]


class EchoServerProtocol(EchoProtocol):
    '''The :class:`EchoProtocol` used by the echo :func:`server`.
    '''
    def response(self, data, rest):
        self.transport.write(data)
        data = data[:-len(self.separator)]
        # If we get a QUIT message, close the transport.
        # Used by the test suite.
        if data == b'QUIT':
            self.transport.close()
        return data


class EchoGreen(AbstractClient):
    """A client for the echo server to be used by the Flask application
    """
    protocol_factory = partial(Connection, EchoProtocol)
    address = None

    def __init__(self, appname, wait, pool_size=5, loop=None):
        super().__init__(loop)
        self.app_name = appname
        self.wait = wait
        self.pool = Pool(self.connect, pool_size, self._loop)

    def connect(self):
        return self.create_connection(self.address)

    def __call__(self, message):
        return self.wait(self._call(message))

    async def _call(self, message):
        # get the address of the echo application
        if not self.address:
            app = await pulsar.get_application(self.app_name)
            self.address = app.cfg.addresses[0]
        connection = await self.pool.connect()
        with connection:
            consumer = connection.current_consumer()
            consumer.start(message)
            await consumer.on_finished
            return consumer.buffer


def FlaskApp(echo):
    app = Flask(__name__)

    @app.errorhandler(404)
    def not_found(e):
        return make_response("404 Page", 404)

    @app.route('/', methods=['GET'])
    def home():
        return "Try any other url for an echo"

    @app.route('/<path>', methods=['GET'])
    def add_org(path):
        a = echo(path.encode('utf-8'))
        return "Flask Url : %s" % a.decode('utf-8')

    return app


class FlaskSite(wsgi.LazyWsgi):
    """Configure a flask_ application to run on a :class:`.GreenPool`
    """
    def setup(self, environ=None):
        green_pool = GreenPool()
        cfg = environ['pulsar.cfg']
        echo_app_name = 'echo_%s' % cfg.name
        echo = EchoGreen(echo_app_name, green_pool.wait)
        app = FlaskApp(echo)
        return wsgi.WsgiHandler([GreenWSGI(app, green_pool)])


class FlaskGreen(MultiApp):
    """Multiapp Server configurator
    """
    cfg = Config(bind=':8080', echo_bind=':8060')

    def build(self):
        yield self.new_app(WSGIServer, callable=FlaskSite())
        yield self.new_app(SocketServer, 'echo', callable=EchoServerProtocol,
                           echo_connection_made=log_connection)


def log_connection(connection, exc=None):
    if not exc:
        connection.logger.debug('Got a new connection to echo server!')


if __name__ == '__main__':  # pragma    nocover
    FlaskGreen().start()



'''Tests the tools and utilities in pulsar.utils.'''
import unittest

from pulsar import system, platform


class TestSystem(unittest.TestCase):

    @unittest.skipUnless(platform.is_posix, 'Posix platform required')
    def testPlatform(self):
        self.assertFalse(platform.is_windows)
        self.assertFalse(platform.is_winNT)

    def test_maxfd(self):
        m = system.get_maxfd()
        self.assertTrue(m)
